#!/usr/bin/env python3.7
from sparcur.config import auth
__doc__ = f"""
SPARC curation cli for fetching, validating datasets, and reporting.
Usage:
    spc clone    [options] <project-id>
    spc pull     [options] [<directory>...]
    spc refresh  [options] [<path>...]
    spc fetch    [options] [<path>...]
    spc find     [options] --name=<PAT>...
    spc status   [options]
    spc meta     [options] [<path>...]
    spc rmeta    [options]
    spc export   [schemas] [options] [<path>...]
    spc report   size    [options] [<path>...]
    spc report   tofetch [options] [<directory>...]
    spc report   terms   [anatomy cells subcelluar] [options]
    spc report   [access filetypes pathids test]    [options]
    spc report   [completeness keywords subjects]   [options]
    spc report   [contributors samples errors mbf]  [options]
    spc report   [(anno-tags <tag>...) changes mis] [options]
    spc report   [(overview [<path>...]) all]       [options]
    spc shell    [affil integration protocols exit] [options]
    spc shell    [(dates [<path>...]) sheets]       [options]
    spc server   [options]
    spc apinat   [options] <path-in> <path-out>
    spc tables   [options] [<directory>...]
    spc annos    [options] [export shell]
    spc feedback <feedback-file> <feedback>...
    spc missing  [options]
    spc xattrs   [options]
    spc demos    [options]
    spc goto     <remote-id>
    spc fix      [options] [duplicates mismatch] [<path>...]
    spc stash    [options --restore] <path>...
    spc make-url [options] [<id-or-path>...]
    spc show     [schemas rmeta (export [json ttl])] [options] [<project-id>]
    spc sheets   [update] [options] <sheet-name>

Commands:
    clone       clone a remote project (creates a new folder in the current directory)

    pull        retrieve remote file structure

                options: --empty
                       : --sparse-limit

    refresh     retrieve remote file sizes and fild ids (can also fetch using the new data)

                options: --fetch
                       : --level
                       : --only-no-file-id

    fetch       fetch remote data based on local metadata (NOTE does NOT refresh first)

                options: --level
                       : --mbf      fetch mbf xml metadata and only for specific datasets

    find        list unfetched files with option to fetch

                options: --name=<PAT>...  glob options should be quoted to avoid expansion
                       : --existing       include existing files in search
                       : --refresh        refresh matching files
                       : --fetch          fetch matching files
                       : --level

    status      list existing files where local meta does not match cached

    meta        display the metadata the current folder or specified paths

                options: --diff     diff the local and cached metadata
                       : --uri      render the uri for the remote
                       : --browser  navigate to the human uri for this file
                       : --human
                       : --context  include context, e.g. dataset

    rmeta       retrieve metadata about files/folders from the remote

    export      export extracted data to json (and everything else)

                schemas         export schemas from python to json

                options: --latest   run derived pipelines from latest json
                       : --partial  run derived pipelines from the latest partial json export
                       : --open=P   open the output file with specified command
                       : --show     open the output file using xopen
                       : --mbf      extract and export mbf embedded metadata

    report      generate reports

                all             generate all reports (use with --to-sheets)
                size            dataset sizes and file counts
                completeness    submission and curation completeness
                filetypes       filetypes used across datasets
                pathids         mapping from local path to cached id
                keywords        keywords used per dataset
                terms           all ontology terms used in the export

                                anatomy
                                cells
                                subcelluar

                subjects        all headings from subjects files
                samples         all headings from samples files
                contributors    report on all contributors
                errors          list of all errors per dataset
                test            do as little as possible (use with --profile)
                mbf             mbf term report (can use with --unique)
                anno-tags       list anno exact for a curation tag
                changes         diff two curation exports
                mis             list summary predicates used per dataset
                access          report on dataset access master vs pipelines
                overview        general dataset information

                options: --raw  run reports on live data without export
                       : --tab-table
                       : --to-sheets
                       : --sort-count-desc
                       : --unique
                       : --uri
                       : --uri-api
                       : --debug
                       : --export-file=PATH
                       : --ttl-file=PATHoURI
                       : --ttl-compare=PATHoURI

    shell       drop into an ipython shell

                integration     integration subshell with different defaults
                exit            (use with --profile)

    server      reporting server

                options: --raw  run server on live data without export

    apinat      convert ApiNATOMY json to rdf and serialize to ttl

    missing     find and fix missing metadata
    xattrs      populate metastore / backup xattrs
    demos       long running example queries
    goto        given an id cd to the containing directory
                invoke as `pushd $(spc goto <id>)`
    dedupe      find and resolve cases with multiple ids
    fix         broke something? put the code to fix it here

                mismatch
                duplicates
    stash       stash a copy of the specific files and their parents
    make-url    return urls for blackfynn dataset ids, or paths

Options:
    -f --fetch              fetch matching files
    -R --refresh            refresh matching files
    -r --rate=HZ            sometimes we can go too fast when fetching [default: 5]
    -l --limit=SIZE_MB      the maximum size to download in megabytes [default: 2]
                            use zero or negative numbers to indicate no limit
    -L --level=LEVEL        how deep to go in a refresh
                            used by any command that acceps <path>...
    -p --pretend            if the defult is to act, dont, opposite of fetch

    -h --human              print human readable values
    -b --browser            open the uri in default browser
    -u --uri                print the human uri for the path in question
    -a --uri-api            print the api uri for the path in question
    -c --context            include context for a file e.g. dataset
    -n --name=<PAT>         filename pattern to match (like find -name)
    -e --empty              only pull empty directories
    -x --exists             when searching include files that have already been pulled
    -m --only-meta          only pull known dataset metadata files
    -z --only-no-file-id    only pull files missing file_id
    -o --overwrite          fetch even if the file exists
    --project-path=<PTH>    set the project path manually
    --sparse-limit=COUNT    package count that forces a sparse pull [default: {auth.get('sparse-limit')}]
                            use zero or negative numbers to indicate no limit

    -F --export-file=PATH   run reports on a specific export file
    -t --tab-table          print simple table using tabs for copying
    -A --latest             run derived pipelines from latest json
    -P --partial            run derived pipelines from the latest partial json export
    -W --raw                run reporting on live data without export
    --to-sheets             push report to google sheets
    --ttl-file=PATHoURI     location of ttl file (uses latest if not specified)
    --ttl-compare=PATHoURI  location of ttl file for comparison

    -S --sort-size-desc     sort by file size, largest first
    -C --sort-count-desc    sort by count, largest first

    -O --open=PROGRAM       open the output file with program
    -w --show               open the output file
    -U --upload             update remote target (e.g. a google sheet) if one exists
    -N --no-google          hack for ipv6 issues
    -D --diff               diff local vs cache

    --port=PORT             server port [default: 7250]

    -j --jobs=N             number of jobs to run             [default: 12]
    -d --debug              drop into a shell after running a step
    -v --verbose            print extra information
    --profile               profile startup performance
    --local                 ignore network issues
    --mbf                   fetch/export mbf related metadata
    --unique                return a unique set of values without additional info

    --log-path=PATH         folder where logs are saved       [default: {auth.get_path('log-path')}]
    --cache-path=PATH       folder where remote data is saved [default: {auth.get_path('cache-path')}]
    --export-path=PATH      base folder for exports           [default: {auth.get_path('export-path')}]
"""

from time import time
start = time()

import os
import sys
import json
import errno
import types
from itertools import chain
from collections import Counter, defaultdict
import idlib
import htmlfn as hfn
import ontquery as oq
start_middle = time()
import augpathlib as aug
from pyontutils import clifun as clif
from pyontutils.core import OntResGit
from pyontutils.utils import UTCNOWISO, subclasses
from pyontutils.config import auth as pauth
from terminaltables import AsciiTable

from sparcur import reports  # top level
from sparcur import datasets as dat
from sparcur import exceptions as exc
from sparcur.core import JT
from sparcur.core import OntId, OntTerm, adops
from sparcur.utils import GetTimeNow  # top level
from sparcur.utils import log, logd, bind_file_handler
from sparcur.paths import Path, BlackfynnCache, StashPath
from sparcur.state import State
from sparcur.backends import BlackfynnRemote
from sparcur.curation import Summary, Integrator
from sparcur.protocols import ProtocolData
from sparcur.blackfynn_api import FakeBFLocal

try:
    breakpoint
except NameError:
    from IPython import embed as breakpoint

stop = time()


class Options(clif.Options):

    @property
    def export_schemas_path(self):
        return Path(self.export_path) / 'schemas'  # FIXME not sure if correct

    @property
    def jobs(self):
        return int(self._args['--jobs'])

    @property
    def limit(self):
        l = int(self._args['--limit'])
        if l >= 0:
            return l

    @property
    def sparse_limit(self):
        l = int(self._args['--sparse-limit'])
        if l >= 0:
            return l

    @property
    def level(self):
        return int(self._args['--level']) if self._args['--level'] else None

    @property
    def rate(self):
        return int(self._args['--rate']) if self._args['--rate'] else None

    @property
    def mbf(self):
        # deal with the fact that both mbf and --mbf are present
        return self._args['--mbf'] or self._default_mbf

    @property
    def show(self):
        # deal with the fact that both show and --show are present
        return self._args['--show'] or self._default_show


class Dispatcher(clif.Dispatcher):
    spcignore = ('.git',
                 '.~lock',)

    def _export(self, export_class, export_source_path=None, org_id=None):

        if export_source_path is None:
            export_source_path = self.cwd

        export_source_path = self.cwd
        export = export_class(self.options.export_path,
                              export_source_path,
                              self._folder_timestamp,
                              self._timestamp,
                              self.options.latest,
                              self.options.partial,
                              (self.options.open
                               if self.options.open else
                               self.options.show),
                              org_id)
        return export

    def _print_table(self, rows, title=None, align=None, ext=None):
        """ ext is only used when self.options.server -> True """
        def asStr(c):
            if hasattr(c, 'asStr'):
                return c.asStr()
            else:
                return str(c)

        def asAsciiRows(rows):
            return [[asStr(cell) for cell in row] for row in rows]

        def simple_tsv(rows):
            return '\n'.join('\t'.join(asAsciiRows(rows))) + '\n'

        if self.options.tab_table:
            if title:
                print(title)

            print(simple_tsv(rows))

        elif self.options.server:
            if ext is not None:
                if ext == '.tsv':
                    nowish = UTCNOWISO('seconds')
                    fn = json.dumps(f'{title} {nowish}')
                    return (
                        simple_tsv(rows), 200,
                        {'Content-Type': 'text/tsv; charset=utf-8',
                         'Content-Disposition': f'attachment; filename={fn}'})
                if isinstance(ext, types.FunctionType):
                    return ext(hfn.render_table(rows[1:], *rows[0]), title=title)
                else:
                    return 'Not found', 404

            return hfn.render_table(rows[1:], *rows[0]), title
        else:
            table = AsciiTable(asAsciiRows(rows), title=title)
            if align:
                assert len(align) == len(rows[0])
                table.justify_columns = {i:('left' if v == 'l'
                                            else ('center' if v == 'c'
                                                  else ('right'
                                                        if v == 'r' else
                                                        'left')))
                                         for i, v in enumerate(align)}
            print(table.table)
            if ext:
                return rows, title

    def _print_paths(self, paths, title=None):
        if self.options.sort_size_desc:
            key = lambda ps: -ps[-1]
        else:
            key = lambda ps: ps

        def derp(p):
            if p.cache is None:
                raise exc.NoCachedMetadataError(p)

        rows = [['Path', 'size', '?'],
                *((p, s.hr
                   if isinstance(s, aug.FileSize) else
                   s, 'x' if p.exists() else '')
                  for p, s in
                  sorted(([p, ('/' if p.is_dir() else
                               (p.cache.meta.size
                                if p.cache.meta.size else
                                '??')
                               if p.cache.meta else '_')]
                          for p in paths if not derp(p)), key=key))]
        self._print_table(rows, title)


class Main(Dispatcher):
    child_port_attrs = ('anchor',
                        'project_path',
                        'project_id',
                        'bfl',
                        'BlackfynnRemote',
                        'summary',
                        'cwd',
                        'cwdintr',
                        '_datasets_with_extension',
                        '_timestamp',
                        '_folder_timestamp',)

    # any attr forced on children must be set before super().__init__ is called
    # set timestamp early so that the loggers can use it
    # I don't think this will cause too much trouble ...
    #_timestamp = None
    #_folder_timestamp = None

    # things all children should have
    # kind of like a non optional provides you WILL have these in your namespace
    def __init__(self, options, time_now=GetTimeNow()):
        self._time_now = time_now
        self._timestamp = self._time_now.START_TIMESTAMP
        self._folder_timestamp = self._time_now.START_TIMESTAMP_LOCAL
        super().__init__(options)
        if not self.options.verbose:
            log.setLevel('INFO')
            logd.setLevel('INFO')

        if self.options.project_path:
            self.cwd = Path(self.options.project_path).resolve()
        else:
            self.cwd = Path.cwd()

        Integrator.rate = self.options.rate
        Integrator.no_google = self.options.no_google

        self.cwdintr = Integrator(self.cwd)

        # pass debug along (sigh)
        from augpathlib import RemotePath, AugmentedPath  # for debug
        AugmentedPath._debug = self.options.debug
        RemotePath._debug = self.options.debug

        # FIXME populate this via decorator
        if (self.options.clone or
            self.options.meta or
            self.options.show or
            self.options.sheets or
            self.options.goto or
            self.options.apinat or
            self.options.tofetch or  # size does need a remote but could do it lazily
            self.options.filetypes or
            self.options.anno_tags or
            self.options.status or  # eventually this should be able to query whether there is new data since the last check
            self.options.pretend or
            (self.options.report and not self.options.raw) or
            (self.options.report and self.options.export_file) or
            (self.options.export and self.options.schemas) or
            (self.options.find and not (self.options.fetch or self.options.refresh))):
            # short circuit since we don't know where we are yet
            Integrator.no_google = True
            return

        elif (self.options.pull or
              self.options.mismatch or
              self.options.stash or
              self.options.contributors or
              self.options.make_url or
              self.options.missing):
            Integrator.no_google = True

        self._setup_local()  # if this isn't run up here the internal state of the program get's wonky

        if (self.options.report and
            not self.options.raw and
            not self.options.access):
            Integrator.setup(local_only=True)  # FIXME sigh
        else:
            try:
                self._setup_bfl()
            except BaseException as e:
                if self.options.local:
                    log.exception(e)
                    self.BlackfynnRemote._api = FakeBFLocal(self.anchor.id, self.anchor)
                    self.BlackfynnRemote.anchorTo(self.anchor)
                else:
                    raise e

        if self.options.export or self.options.shell:
            self._setup_export()
            self._setup_ontquery()

    def _setup_local(self):
        self.BlackfynnRemote = BlackfynnCache._remote_class
        self.BlackfynnRemote._async_rate = self.options.rate

        local = self.cwd

        # we have to start from the cache class so that
        # we can configure
        try:
            _cpath = local.cache  # FIXME project vs subfolder
            if _cpath is None:
                raise exc.NoCachedMetadataError  # FIXME somehow we decided not to raise this!??!

            self.anchor = _cpath.anchor
        except exc.NoCachedMetadataError as e:
            root = local.find_cache_root()
            if root is not None:
                self.anchor = root.cache
                if local.skip_cache:
                    print(f'{local} is ignored!')
                    sys.exit(112)

                raise NotImplementedError('TODO recover meta?')
            else:
                print(f'{local} is not in a project!')
                sys.exit(111)

        # replace a bottom up anchoring rule with top down
        self.anchor.anchorClassHere(remote_init=False)

        self.project_path = self.anchor.local
        self.summary = Summary(self.project_path)
        Summary._n_jobs = self.options.jobs
        if self.options.debug:
            Summary._debug = True

    def _setup_bfl(self):
        self.BlackfynnRemote._setup()
        self.BlackfynnRemote.init(self.anchor.id)

        self.bfl = self.BlackfynnRemote._api
        State.bind_blackfynn(self.bfl)

    def _setup_export(self):
        Integrator.setup()

    def _setup_ontquery(self):
        # FIXME this should be in its own setup method
        # pull in additional graphs for query that aren't loaded properly
        RDFL = oq.plugin.get('rdflib')
        olr = Path(pauth.get_path('ontology-local-repo'))
        branch = 'methods'
        for fn in ('methods', 'methods-helper', 'methods-core'):
            org = OntResGit(olr / f'ttl/{fn}.ttl', ref=branch)
            OntTerm.query.ladd(RDFL(org.graph, OntId))

    @property
    def project_name(self):
        return self.anchor.name
        #return self.bfl.organization.name

    @property
    def project_id(self):
        #self.bfl.organization.id
        return self.anchor.id

    @property
    def datasets(self):
        # XXX DO NOT YIELD DIRECTLY FROM self.anchor.children
        # unless you are cloning or something like that
        # because cache.children completely ignores existing
        # files and folders and there really isn't a safe way
        # to use it once files already exist because then
        # viewing the cached children would move all the folders
        # around, file under sigh, yes fix CachePath construction
        for local in self.datasets_local:
            yield local.cache

        #yield from self.anchor.children  # NOT OK TO YIELD FROM THIS URG

    @property
    def datasets_remote(self):
        for d in self.anchor.remote.children:
            # FIXME lo the crossover (good for testing assumptions ...)
            #yield d.local
            yield d

    @property
    def datasets_local(self):
        for d in self.anchor.local.children: #self.datasets:
            if d.exists():
                yield d

    ###
    ## vars
    ###

    @property
    def directories(self):
        return [Path(string_path).absolute()
                for string_path in self.options.directory]

    @property
    def paths(self):
        return [Path(string_path).absolute()
                for string_path in self.options.path]

    @property
    def _paths(self):
        """ all relevant paths determined by the flags that have been set """
        # but if you use the generator version of _paths
        # then if you add a folder to the previous path
        # then it will yeild that folder! which is SUPER COOL
        # but breaks lots of asusmptions elsehwere
        paths = self.paths
        if not paths:
            paths = self.cwd,  # don't call Path.cwd() because it may have been set from --project-path

        if self.options.only_meta:
            paths = (mp.absolute()
                     for p in paths
                     for mp in dat.DatasetStructureLax(p).meta_paths)
            yield from paths
            return

        yield from self._build_paths(paths)

    def _build_paths(self, paths):
        def inner(paths, level=0, stop=self.options.level):
            """ depth first traversal of children """
            for path in paths:
                if self.options.only_no_file_id:
                    if (path.is_broken_symlink() and
                        (path.cache.meta.file_id is None)):
                        yield path
                        continue

                elif self.options.empty:
                    if path.is_dir():
                        try:
                            next(path.children)
                            # if a path has children we still want to
                            # for empties in them to the level specified
                        except StopIteration:
                            yield path
                    else:
                        continue
                else:
                    yield path

                if stop is None:
                    if self.options.only_no_file_id:
                        for rc in path.rchildren:
                            if (rc.is_broken_symlink() and
                                rc.cache.meta.file_id is None):
                                yield rc
                    else:
                        yield from path.rchildren

                elif level <= stop:
                    yield from inner(path.children, level + 1)

        yield from inner(paths)

    @property
    def _dirs(self):
        for p in self._paths:
            if p.is_dir():
                yield p

    @property
    def _not_dirs(self):
        for p in self._paths:
            if not p.is_dir():
                yield p

    def clone(self):
        project_id = self.options.project_id
        if project_id is None:
            print('no remote project id listed')
            sys.exit(4)
        # given that we are cloning it makes sense to _not_ catch a connection error here
        self.BlackfynnRemote = BlackfynnRemote._new(Path, BlackfynnCache)
        try:
            self.BlackfynnRemote.init(project_id)
        except exc.MissingSecretError:
            print(f'missing api secret entry for {project_id}')
            sys.exit(11)

        # make sure that we aren't in a project already
        existing_root = self.cwd.find_cache_root()
        if existing_root is not None and existing_root != self.cwd:
            message = f'fatal: already in project located at {existing_root.resolve()!r}'
            print(message)
            sys.exit(3)

        try:
            if self.options.project_path:
                anchor = self.BlackfynnRemote.anchorTo(self.cwd, create=True)
            else:
                anchor = self.BlackfynnRemote.dropAnchor(self.cwd)
        except exc.DirectoryNotEmptyError:
            message = (f'fatal: destination path already '
                       'exists and is not an empty directory.')
            print(message)
            sys.exit(2)
        except BaseException as e:
            log.exception(e)
            sys.exit(11111)

        anchor.local_data_dir_init()

        self.anchor = anchor
        self.project_path = self.anchor.local
        with anchor:
            # update self.cwd so pull sees the right thing
            self.cwd = Path.cwd()
            self._clone_pull()

    def _clone_pull(self):
        """ wow this really shows that the pull -> bootstrap workflow
            is completely broken and stupidly slow """

        # populate just the dataset folders
        datasets = list(self.anchor.children)
        # populate the dataset metadata
        self.rmeta(use_cache_path=True, exist_ok=True)

        # mark sparse datasets so that we don't have to
        # fiddle with detecting sparseness during bootstrap
        sparse_limit = self.options.sparse_limit
        [d._sparse_materialize(sparse_limit=sparse_limit) for d in datasets]
        sparse = [d for d in datasets if d.is_sparse()]  # sanity check


        skip = auth.get_list('datasets-no')  # FIXME this should be materialized as well

        # pull all the files
        self.pull()

    def _total_package_counts(self):
        from sparcur.datasources import BlackfynnDatasetData

        def total_packages(cache):
            bdd = BlackfynnDatasetData(cache)
            try:
                blob = bdd.fromCache()
            except FileNotFoundError as e:
                # FIXME TODO also check cached rmeta dates during pull
                log.warning(e)
                blob = bdd()

            return sum(blob['package_counts'].values())

        totals = []
        for cache in self.datasets:
            totals.append((cache, total_packages(cache)))

        return totals

    def _sparse_from_metadata(self, sparse_limit:int=None):
        if sparse_limit is None:
            sparse_limit = self.options.sparse_limit

        if sparse_limit is None:
            return []

        totals = self._total_package_counts()
        #tps = sorted([ for d in self.datasets], key= lambda ab: ab[1])
        sparse = [r.id for r, pc in totals if pc >= sparse_limit]
        return sparse

    def pull(self):
        # FIXME this still isn't quite right because it does not correctly
        # deal dataset renaming, but it is much closer
        from joblib import Parallel, delayed
        dirs = self.directories
        dirs = dirs if dirs else (self.cwd,)
        for d in dirs:
            d.pull(
                time_now=self._time_now,
                debug=self.options.debug,
                n_jobs=12,
                log_level='DEBUG' if self.options.verbose else 'INFO',
                Parallel=Parallel,
                delayed=delayed,)

    def refresh(self):
        paths = self.paths
        cwd = self.cwd
        if not paths:
            paths = cwd,

        to_root = sorted(set(parent
                             for path in paths
                             for parent in path.parents
                             if parent.cache is not None),
                         key=lambda p: len(p.parts))

        if self.options.pretend:
            ap = list(chain(to_root, self._paths))
            self._print_paths(ap)
            print(f'total = {len(ap):<10}rate = {self.options.rate}')
            return

        self._print_paths(chain(to_root, self._paths))

        from pyontutils.utils import Async, deferred
        hz = self.options.rate
        fetch = self.options.fetch
        limit = self.options.limit

        drs = [d.remote for d in chain(to_root, self._dirs)]

        if not self.options.debug:
            refreshed = Async(rate=hz)(deferred(r.refresh)(
                update_data_on_cache=r.cache.is_file() and
                r.cache.exists())
                                       for r in drs)
        else:
            refreshed = [r.refresh(update_data_on_cache=r.cache.is_file() and
                                   r.cache.exists()) for r in drs]

        moved = []
        parent_moved = []
        for new, r in zip(refreshed, drs):
            if new is None:
                log.critical('utoh')

            oldl = r.local
            try:
                r.update_cache()  # calling this directly is ok for directories
            except FileNotFoundError as e:
                parent_moved.append(oldl)
                continue
            except OSError as e:
                if e.errno == errno.ENOTEMPTY:
                    log.error(f'{e}')
                    continue
                else:
                    raise e

            newl = r.local
            if oldl != newl:
                moved.append([oldl, newl])

        if moved:
            self._print_table(moved, title='Folders moved')
            for old, new in moved:
                if old == cwd:
                    log.info(f'Changing directory to {new}')
                    new.chdir()

        if parent_moved:
            self._print_paths(parent_moved, title='Parent moved')

        if not self.options.debug:
            refreshed = Async(rate=hz)(deferred(path.cache.refresh)(
                update_data=fetch, size_limit_mb=limit)
                                       for path in self._not_dirs)

        else:
            for path in self._not_dirs:
                path.cache.refresh(update_data=fetch, size_limit_mb=limit)

    def _datasets_with_extension(self, extension):
        """ Hack around the absurd slowness of python's rglob """

        command = fr"""for d in */; do
    find "$d" \( -type l -o -type f \) -name '*.{extension}' \
    -exec getfattr -n user.bf.id --only-values "$d" \; -printf '\n' -quit ;
done"""
        with os.popen(command) as p:
            string = p.read()

        has_extension = string.split('\n')
        datasets = [d for d in self.datasets if d.id in has_extension]
        return datasets

    def _fetch_mbf(self):
        from pyontutils.utils import Async, deferred
        def fetch_mbf_metadata(dataset):
            xmls = dataset.rglob('*.xml')
            Async(rate=5)(deferred(x.fetch)(size_limit_mb=None)
                          for x in xmls if not x.exists())

        for dataset in self._datasets_with_extension('xml'):
            fetch_mbf_metadata(dataset)

    def fetch(self):
        if self.options.mbf:  # FIXME hack -> fetch '*.xml' or something
            self._fetch_mbf()
            return

        paths = [p for p in self._paths if not p.is_dir()]
        self._print_paths(paths)
        if self.options.pretend:
            return

        from pyontutils.utils import Async, deferred
        hz = self.options.rate
        Async(rate=hz)(deferred(path.cache.fetch)(
            size_limit_mb=self.options.limit)
                       for path in paths
                       if not path.exists()
                       # FIXME need a staging area ...
                       # FIXME also, the fact that we sometimes need content_different
                       # means that there may be silent fetch failures
                       or path.content_different())

    def _check_duplicates(self, datasets):
        # NOTE this is an ok sanity check
        # but cannot catch the duplicate dataset issue
        # which actually comes from calling refresh
        # while iterating through the list of dataset paths
        # which can and does cause datasets move due to
        # changes in their folder name
        if not datasets:
            datasets = list(self.datasets_local)

        counts = Counter([d['id'] if isinstance(d, dict) else d.id
                          for d in datasets])
        bads = [(id, c) for id, c in counts.most_common() if c > 1]
        if bads:
            # FIXME sys.exit ?
            raise BaseException(f'duplicate datasets {bads}')

    def _check_exists(self, dataset_paths):
        bads = [p for p in dataset_paths if not p.exists()]
        if bads:
            raise FileNotFoundError(f'The following do not exist!\n{bads}')

    def export(self):
        from sparcur import export as ex  # FIXME very slow to import
        from sparcur import schemas as sc

        # FIXME export should be able to run without needing any external
        # data the fetch step should happen before the export so that
        # network connections don't creep into the process

        # TODO export source path without having to be in that folder

        if self.options.schemas:
            ex.core.export_schemas(self.options.export_schemas_path)

        elif self.options.mbf:
            export = self._export(ex.ExportXml)
            dataset_paths = self._datasets_with_extension('xml')
            self._check_duplicates(dataset_paths)
            self._check_exists(dataset_paths)
            blob_ir, *rest = export.export(dataset_paths=dataset_paths,
                                           jobs=self.options.jobs,
                                           debug=self.options.debug)

            return blob_ir

        else:
            export = self._export(ex.Export)
            dataset_paths = tuple(self.paths)
            self._check_duplicates(dataset_paths)  # NOTE can be empty
            self._check_exists(dataset_paths)
            noexport = (auth.get_list('datasets-noexport') +
                        auth.get_list('datasets-no'))
            blob_ir, *rest = export.export(dataset_paths=dataset_paths,
                                           exclude=noexport)

            sc.SummarySchema().validate_strict(export.latest_export)

        if self.options.debug:
            breakpoint()

    def annos(self):
        if self.options.raw:
            data = self.summary.data()
        else:
            from sparcur import export as ex
            data = self._export(ex.Export).latest_ir

        dataset_blobs = data['datasets']

        from protcur.analysis import protc
        from sparcur.protocols import ProtcurData
        ProtcurData.populate_annos()

        class ProtocolActual(ProtocolData):  # FIXME so ... bad ...
            @property
            def protocol_uris(self, outer_self=self):  # FIXME this needs to be pipelined
                for d in dataset_blobs:
                    try:
                        yield from adops.get(d, ['meta', 'protocol_url_or_doi'])
                    except exc.NoSourcePathError:
                        pass

        if self.options.export:
            with open('/tmp/sparc-protcur.rkt', 'wt') as f:
                f.write(protc.parsed())

        pa = ProtocolActual()
        all_blackfynn_uris = set(pa.protocol_uris_resolved)
        all_hypothesis_uris = set(a.uri for a in protc)
        if self.options.shell or self.options.debug:
            p, *rest = self._paths
            f = Integrator(p)
            all_annos = [list(protc.byIri(uri))
                         for uri in f.protocol_uris_resolved]
            breakpoint()

    def demos(self):
        # get the first dataset
        dataset = next(iter(self.summary))

        # another way to get the first dataset
        dataset_alt = next(self.anchor.children)

        # view all dataset descriptions call repr(tabular_view_demo)
        ds = self.datasets
        tabular_view_demo = [next(d.dataset_description).t
                             for d in ds[:1]
                             if 'dataset_description' in d.data]  # FIXME

        # get package testing
        bigskip = auth.get_list('datasets-sparse') + auth.get_list('datasets-no')
        bfds = self.bfl.bf.datasets()
        packages = [list(d.packages) for d in bfds[:3]
                    if d.id not in bigskip]
        n_packages = [len(ps) for ps in packages]

        # bootstrap a new local mirror
        # FIXME at the moment we can only have of these at a time
        # sigh more factories incoming
        #anchor = BlackfynnCache('/tmp/demo-local-storage')
        #anchor.bootstrap()

        if False:
            ### this is the equivalent of export, quite slow to run
            # export everything
            dowe = self.summary.data()

            # show all the errors from export everything
            error_id_messages = [(d['id'], e['message'])
                                 for d in dowe['datasets']
                                 for e in d['errors']]
            error_messages = [e['message']
                              for d in dowe['datasets']
                              for e in d['errors']]

        #rchilds = list(datasets[0].rchildren)
        #package, file = [a for a in rchilds if a.id == 'N:package:8303b979-290d-4e31-abe5-26a4d30734b4']

        return self.shell()

    def tables(self):
        """ print summary view of raw metadata tables, possibly per dataset """

        dat.DatasetStructure._refresh_on_missing = False
        dat.SubmissionFile._refresh_on_missing = False
        dat.SubjectsFile._refresh_on_missing = False
        dat.SamplesFile._refresh_on_missing = False
        directories = self.directories
        directories = directories if directories else (self.cwd,)
        for directory in directories:
            self._tables_dir(directory)

    def _tables_dir(self, directory):
        droot = dat.DatasetStructure(directory)
        if droot.cache.is_dataset():
            datasetdatas = droot,
        elif droot.cache.is_organization():
            datasetdatas = droot.children
        else:
            raise TypeError(f'whats a {type(droot)}? {droot}')

        def do_file(d, ft):
            if hasattr(d, ft):
                f = getattr(d, ft)
                fs = f if isinstance(f, list) else [f]
                for f in fs:
                    if hasattr(f, 'object'):
                        yield (f.object._t())
                    else:
                        message = f'object missing on {ft} for {d}'
                        log.warning(message)
                        yield message

            else:
                yield f'{ft} missing on {d}'

        tables = [t for d in datasetdatas for t in
                  (*do_file(d, 'dataset_description'),
                   *do_file(d, 'submission'),
                   *do_file(d, 'subjects'),
                   *do_file(d, 'samples'),
                   *do_file(d, 'manifest'),
                   ('\n--------------------------------------------------\n'
                      '=================================================='
                    '\n--------------------------------------------------\n'))]

        [print(t if isinstance(t, str) else repr(t)) for t in tables]  # FIXME _print_tables?

    def find(self):
        paths = []
        if self.options.name:  # has to always be true now
            patterns = self.options.name
            path = self.cwd
            for pattern in patterns:
                # TODO filesize mismatches on non-fake
                # no longer needed due to switching to symlinks
                #if '.fake' not in pattern and not self.options.overwrite:
                    #pattern = pattern + '.fake*'

                for file in path.rglob(pattern):
                    paths.append(file)

        if paths:
            paths = [p for p in paths if not p.is_dir()]
            search_exists = self.options.exists
            if self.options.limit:
                old_paths = paths
                paths = [p for p in paths
                         if p.cache.meta.size is None or  # if we have no known size don't limit it
                         search_exists or
                         not p.exists() and
                         p.cache.meta.size.mb < self.options.limit or
                         p.exists() and p.size != p.cache.meta.size and
                         (not log.info(f'Truncated transfer detected for {p}\n'
                                       f'{p.size} != {p.cache.meta.size}'))
                         and p.cache.meta.size.mb < self.options.limit]

                n_skipped = len(set(p for p in old_paths
                                    if p.is_broken_symlink()) - set(paths))

            if self.options.pretend:
                self._print_paths(paths)
                print(f'skipped = {n_skipped:<10}rate = {self.options.rate}')
                return

            if self.options.verbose:
                for p in paths:
                    print(p.cache.meta.as_pretty(pathobject=p))

            if self.options.fetch or self.options.refresh:
                from pyontutils.utils import Async, deferred
                hz = self.options.rate  # was 30
                limit = self.options.limit
                fetch = self.options.fetch
                if self.options.refresh:
                    Async(rate=hz)(deferred(path.remote.refresh)(
                        update_cache=True, update_data=fetch, size_limit_mb=limit)
                                   for path in paths)
                elif fetch:
                    Async(rate=hz)(deferred(path.cache.fetch)(
                        size_limit_mb=limit)
                                   for path in paths)

            else:
                self._print_paths(paths)
                print(f'skipped = {n_skipped:<10}rate = {self.options.rate}')

    def feedback(self):
        file = self.options.feedback_file
        feedback = ' '.join(self.options.feedback)
        path = Path(file).resolve()
        eff = Integrator(path)
        # TODO pagenote and/or database
        print(eff, feedback)

    def missing(self):
        for path in self._paths:
            for rc in path.rchildren:
                if rc.is_broken_symlink():
                    m = rc.cache.meta
                    if m.file_id is None:
                        #print(rc)
                        print(m.as_pretty(pathobject=rc))
        #self.bfl.find_missing_meta()

    def xattrs(self):
        raise NotImplementedError('This used to populate the metastore, '
                                  'but currently does nothing.')

    def meta(self):
        if self.options.browser:
            import webbrowser

        BlackfynnCache._local_class = Path  # since we skipped _setup
        Path._cache_class = BlackfynnCache
        paths = self.paths
        if not paths:
            paths = self.cwd,

        old_level = log.level
        if not self.options.verbose:
            log.setLevel('ERROR')
        def inner(path):
            if self.options.uri or self.options.browser:
                uri = path.cache.uri_human
                print('+' + '-' * (len(uri) + 2) + '+')
                print(f'| {uri} |')
                if self.options.browser:
                    webbrowser.open(uri)

            try:
                cmeta = path.cache.meta
                if cmeta is not None:
                    if self.options.diff:
                        if cmeta.checksum is None:
                            if not path.cache.local_object_cache_path.exists():
                                # we are going to have to go to network
                                self._setup()

                            cmeta.checksum = path.cache.checksum()  # FIXME :/
                        lmeta = path.meta
                        # id and file_id are fake in this instance
                        setattr(lmeta, 'id', None)
                        setattr(lmeta, 'file_id', None)
                        print(lmeta.as_pretty_diff(cmeta, pathobject=path,
                                                   human=self.options.human))
                    else:
                        print(cmeta.as_pretty(pathobject=path,
                                              human=self.options.human))

                    if self.options.context:
                        pc = path.cache
                        if not pc.is_organization():
                            org = pc.organization.local
                            if not pc.is_dataset():
                                ds = pc.dataset.local
                                inner(ds)
                            elif self.options.verbose:
                                inner(org)

                        print()

            except exc.NoCachedMetadataError:
                print(f'No metadata for {path}. Run `spc refresh {path}`')

        for path in paths:
            inner(path)

        log.setLevel(old_level)

    def goto(self):
        # TODO this needs an inverted index
        if self.options.remote_id.startswith('N:dataset:'):
            gen = self.cwd.children
        else:
            gen = self.cwd.rchildren

        for rc in gen:
            try:
                if rc.cache.id == self.options.remote_id:
                    if rc.is_broken_symlink() or rc.is_file():
                        rc = rc.parent

                    print(rc.relative_to(self.cwd).as_posix())
                    return
            except AttributeError as e:
                if not rc.skip_cache:
                    log.critical(rc)
                    log.error(e)

    def status(self):
        project_path = self.cwd.find_cache_root()
        if project_path is None:
            print(f'{self.cwd} is not in a project!')
            sys.exit(111)

        existing_files = [f for f in project_path.rchildren if f.is_file()]
        different = []
        for f in existing_files:
            try:
                cmeta = f.cache.meta
            except AttributeError:
                if f.skip_cache:
                    continue

            lmeta = f.meta if cmeta.checksum else f.meta_no_checksum
            # id and file_id are fake in this instance
            setattr(lmeta, 'id', None)
            setattr(lmeta, 'file_id', None)
            if lmeta.content_different(cmeta):
                if self.options.status:
                    print(lmeta.as_pretty_diff(cmeta, pathobject=f,
                                               human=self.options.human))
                else:
                    yield f, lmeta, cmeta

    def server(self):
        from sparcur.server import make_app
        if self.options.raw:
            # FIXME ...
            self.dataset_index = {d.meta.id:Integrator(d)
                                  for d in self.datasets}
        else:
            from sparcur import export as ex
            data = self._export(ex.Export).latest_ir
            self.dataset_index = {d['id']:d for d in data['datasets']}

        report = Report(self)
        app, *_ = make_app(report, self.project_path)
        self.app = app  # debug only
        app.debug = False
        app.run(host='localhost', port=self.options.port, threaded=True)

    def stash(self, paths=None, stashmetafunc=lambda v:v):
        if paths is None:
            paths = self.paths

        stash_base = self.anchor.local.parent / 'stash'  # TODO move to operations
        to_stash = sorted(set(parent for p in paths for parent in
                               chain((p,), p.relative_to(self.anchor).parents)),
                           key=lambda p: len(p.as_posix()))

        if self.options.restore:
            # horribly inefficient, maybe build on a default dict?
            rcs = sorted((c for c in stash_base.rchildren if not c.is_dir()),
                         key=lambda p:p.as_posix(), reverse=True)
            for path in paths:
                relpath = path.relative_to(self.anchor)
                for p in rcs:
                    if p.parts[-len(relpath.parts):] == relpath.parts:
                        p.copy_to(path, force=True, copy_cache_meta=True)  # FIXME old remote may have been deleted, worth a check?
                        # TODO checksum? sync?
                        break

            breakpoint()

        else:
            stash = StashPath(stash_base, self._folder_timestamp)
            new_anchor = stash / self.anchor.name
            new_anchor.mkdir(parents=True)
            new_anchor.cache_init(self.anchor.meta, anchor=True)
            for p in to_stash[1:]:
                p = p.relative_to(self.anchor) if p.root == '/' else p
                new_path = new_anchor / p
                log.debug(p)
                log.debug(new_path)
                p = self.anchor.local / p
                if p.is_dir():
                    new_path.mkdir()
                    new_path.cache_init(p.cache.meta)
                else:
                    # TODO search existing stashes to see if
                    # we already have a stash of the file
                    log.debug(f'{p!r} {new_path!r}')
                    new_path.copy_from(p)
                    pc, npc = p.checksum(), new_path.checksum()
                    # TODO a better way to do this might be to
                    # treat the stash as another local for which
                    # the current local is the remote
                    # however this might require layering
                    # remote and remote remote metadata ...
                    assert pc == npc, f'\n{pc}\n{npc}'
                    cmeta = p.cache.meta
                    nmeta = stashmetafunc(cmeta)
                    log.debug(nmeta)
                    new_path.cache_init(nmeta)

            nall = list(new_anchor.rchildren)
            return nall

    def apinat(self):
        path_in = Path(self.options.path_in)
        path_out = Path(self.options.path_out)
        with open(path_in) as f:
            resource_map = json.load(f)

        if path_in.suffix == '.json':
            from sparcur import apinat
            agraph = apinat.Graph(resource_map)
            graph = agraph.graph()
            graph.write(path=path_out)
        elif path_in.suffix == '.jsonld':
            from pyontutils.core import populateFromJsonLd, OntGraph
            g = populateFromJsonLd(OntGraph(), path_in).write(path=path_out)

    def rmeta(self, use_cache_path=False, exist_ok=False):
        from pyontutils.utils import Async, deferred
        from sparcur.datasources import BlackfynnDatasetData
        dsr = self.datasets if use_cache_path else self.datasets_remote
        all_ = [BlackfynnDatasetData(r) for r in dsr]
        prepared = [bdd for bdd in all_ if not (exist_ok and bdd.cache_path.exists())]
        hz = self.options.rate
        if not self.options.debug:
            blobs = Async(rate=hz)(deferred(d)() for d in prepared)
        else:
            blobs = [d() for d in  prepared]

    def make_url(self):
        lu = {d.cache.id:d.cache for d in self.datasets_local}
        for iop in self.options.id_or_path:
            if iop.startswith('N:dataset:'):
                cache = lu[iop]
            else:
                cache = Path(iop).cache

            print(hfn.atag(cache.uri_human, cache.id))

    def show(self):
        if self.options.export:
            from sparcur import export as ex
            org_id = self.options.project_id if self.options.project_id else auth.get('blackfynn-organization')
            export = self._export(ex.Export, org_id=org_id)
            if self.options.json:
                export.latest_export_path.xopen(self.options.open)
            elif self.options.ttl:
                export.latest_ttl_path.xopen(self.options.open)

        elif self.options.schemas:
            print(self.options.export_schemas_path)
            self.options.export_schemas_path.xopen(self.options.open)  # NOTE it is a folder

        elif self.options.rmeta:
            from sparcur import datasources as ds
            Path(ds.BlackfynnDatasetData.cache_base).xopen(self.options.open)

    def sheets(self):
        from pyontutils import sheets as ps

        # get ir
        if self.options.raw:
            data = self.summary.data()
        else:
            from sparcur import export as ex
            org_id = auth.get('blackfynn-organization')
            export = self._export(ex.Export, org_id=org_id)
            data = export.latest_ir

        # check that the ir is sane
        self._check_duplicates(data['datasets'])

        # get sheets
        sheets = {(s.name, s.sheet_name):s for s in subclasses(ps.Sheet)
                  if not s.__name__.startswith('_')}
        by_sn = {sn:s for (n, sn), s in sheets.items()}
        sn = self.options.sheet_name
        Sheet = by_sn[sn]
        Sheet.fetch_grid = False

        if self.options.update:
            sheet = Sheet(readonly=False)
            sheet.update_from_ir(data)
        else:
            sheet = Sheet()

        if sheet.sheet_name == 'Organs':
            h = sheet.row_object(0)
            tech_cells = [
                cell
                for hcell in h.cells if 'technique' in hcell.value
                for cell in hcell.column.cells[1:] if cell.value]
            sheet.fetch_grid = True
            sheet.fetch()

        if self.options.debug:
            breakpoint()

        return sheet

    ### sub dispatchers

    def report(self):
        report = Report(self)
        return report('report')

    def shell(self):
        """ drop into an shell with classes loaded """
        shell = Shell(self)
        return shell('shell')

    def fix(self):
        fix = Fix(self)
        return fix('fix')


class Report(reports.Report, Dispatcher):

    paths = Main.paths
    _paths = Main._paths

    _print_table = Main._print_table


class Shell(Dispatcher):
    # property ports
    paths = Main.paths
    _paths = Main._paths
    _build_paths = Main._build_paths
    datasets = Main.datasets
    datasets_local = Main.datasets_local
    stash = Main.stash

    def exit(self):
        """ useful for profiling startup time issues """
        print('Peace.')
        return

    def sheets(self):
        from sparcur import sheets
        wev = sheets.WorkingExecVerb()
        wev.condense()
        organs = sheets.Organs(readonly=True)
        breakpoint()

    def default(self):
        datasets = list(self.datasets)
        datas = [Integrator(d).datasetdata for d in datasets]
        datasets_local = list(self.datasets_local)
        dsd = {d.meta.id:d for d in datasets}
        ds = datasets
        summary = self.summary
        org = Integrator(self.project_path)

        asdf = []
        for d in datasets:
            asdf.append(list(d.children))

        breakpoint()

    def dates(self):
        paths = self.paths
        paths = paths if paths else list(self.datasets_local)
        asdf = {p:p.updated_cache_transitive() for p in paths}
        s = sorted(asdf.items(),
                   key=lambda kv: (1, kv[0]) if kv[-1] is None else (0, kv[-1]),
                   reverse=True)
        breakpoint()

    def more(self):
        p, *rest = self._paths
        if p.cache.is_dataset():
            intr = Integrator(p)
            j = JT(intr.data())
            iddf = intr.datasetdata.dataset_description.object  # finally
            #triples = list(f.triples)

        try:
            latest_datasets = self._export(None).latest_export['datasets']
        except:
            pass

        datasets = self.datasets
        datas = [Integrator(d).datasetdata for d in datasets]
        dsd = {d.meta.id:d for d in datasets}

        rcs = list(datasets[-1].rchildren)
        asdf = rcs[-1]
        urg = list(asdf.data)
        resp = asdf.data_headers

        did = self.cwd.cache.id
        r = self.cwd.remote
        if did in dsd:
            dsi = datas.index(dsd[did])
        else:
            dsi = 2

        d = datas[dsi]
        try:
            smf = d.submission.object
            smd = smf.data

            ddf = d.dataset_description.object
            ddd = ddf.data

            suf = d.subjects.object
            sud = suf.data

            saf = d.samples.object
            sad = saf.data

        except AttributeError as e:
            logd.error(f'{d} is missing some file')

        breakpoint()

    def affil(self):
        from pyontutils.utils import Async, deferred
        from sparcur import sheets
        a = sheets.Affiliations()
        m = a.mapping
        rors = sorted(set(_ for _ in m.values() if _))
        #dat = Async(rate=5)(deferred(lambda r:r.data)(i) for i in rors)
        dat = [r.data for r in rors]  # once the cache has been populated
        breakpoint()

    def protocols(self):
        """ test protocol identifier functionality """
        org = Integrator(self.project_path)
        from pyontutils.utils import Async, deferred
        skip = '"none"', 'NA', 'no protocols', 'take protocol from other spreadsheet, ', 'na'
        asdf = [us for us in sorted(org.organs_sheet.byCol.protocol_url_1)
                if us not in skip and us and ',' not in us]
        wat = [idlib.Auto(_) for _ in asdf]
        inst = [i for i in wat]
        res = Async(rate=5)(deferred(i.dereference)(idlib.Auto) for i in inst)
        pis = [i for i in res if isinstance(i, idlib.Pio)]
        #dat = Async(rate=5)(deferred(lambda p: p.data)(i) for i in pis)
        #dois = [d['protocol']['doi'] for d in dat if d]
        dois = [p.doi for p in pis]
        breakpoint()

    def integration(self):
        from protcur.analysis import protc, Hybrid
        #from sparcur.sheets import Organs, Progress, Grants, ISAN, Participants, Protocols as ProtocolsSheet
        from sparcur.protocols import ProtcurData
        p, *rest = self._paths
        intr = Integrator(p)
        j = JT(intr.data)
        pj = list(intr.protocol_jsons)
        pc = list(intr.triples_exporter.protcur)
        #apj = [pj for c in intr.anchor.children for pj in c.protocol_jsons]
        breakpoint()


class Fix(Shell):

    def default(self):
        pass

    def mismatch(self):
        """ once upon a time it was (still at the time of writing) possible
            to update meta on an existing file without preserving the old data
            AND without downloading the new data (insanity) this can be used to fix
            those cases, preserving the old version """
        oops = list(self.parent.status())
        [print(lmeta.as_pretty_diff(cmeta, pathobject=path, human=self.options.human))
         for path, lmeta, cmeta in oops]
        paths = [p for p, *_ in oops]

        def sf(cmeta):
            nmeta = aug.PathMeta(id=cmeta.old_id)
            assert nmeta.id, f'No old_id for {cmeta}'
            return nmeta

        nall = self.stash(paths, stashmetafunc=sf)
        [print(n.cache.meta.as_pretty(n)) for n in nall]
        breakpoint()
        # once everything is in order and backed up 
        # [p.cache.fetch() for p in paths]

    def duplicates(self):
        all_ = defaultdict(list)
        if not self.options.path:
            paths = self.anchor.local.rchildren
        else:
            paths = self.paths

        for rc in paths:
            if rc.cache is None:
                if not rc.skip_cache:
                    log.critical(f'WHAT THE WHAT {rc}')

                continue

            all_[rc.cache.id].append(rc)

        def mkey(p):
            mns = p.cache.meta
            return (not bool(mns),
                    not bool(mns.updated),
                    -mns.updated.timestamp())

        dupes = {i:sorted(l, key=mkey)#, reverse=True)
                 for i, l in all_.items() if len(l) > 1}
        dv = list(dupes.values())
        deduped = [a.dedupe(b, pretend=True) for a, b, *c in dv
                   if (not log.warning(c) if c else not c)
        ]  # FIXME assumes a single dupe...
        to_remove = [d for paths, new in zip(dv, deduped)
                     for d in paths if d != new]
        to_remove_size = [p for p in to_remove if p.cache.meta.size is not None]
        #[p.unlink() for p in to_remove if p.cache.meta.size is None] 
        breakpoint()


def main():
    time_now = GetTimeNow()
    from docopt import docopt, parse_defaults
    args = docopt(__doc__, version='spc 0.0.0')
    defaults = {o.name:o.value if o.argcount else None for o in parse_defaults(__doc__)}

    logpath = Path(args['--log-path'])
    if not logpath.exists():
        logpath.mkdir(parents=True)

    try:
        logfile = logpath / time_now.START_TIMESTAMP_SAFE  # FIXME configure and switch
        bind_file_handler(logfile)

        options = Options(args, defaults)
        main = Main(options, time_now)
        if main.options.debug:
            print(main.options)

        main()
    except BaseException as e:
        log.exception(e)
        print()
        raise e
    finally:
        if logfile.size == 0:
            logfile.unlink()

    if options.profile:
        exit = time()
        a = start_middle - start
        b = (stop - start_middle)
        c = (stop - start)
        d = (exit - stop)
        e = (exit - start)
        s = f'{a} + {b} = {c}\n{c} + {d} = {e}'
        print(s)


if __name__ == '__main__':
    main()



import copy
import json
import itertools
from pathlib import PurePath
from datetime import datetime
from functools import wraps
from collections import deque, defaultdict
import idlib
import rdflib
import ontquery as oq
from idlib.formats import rdf as _bind_rdf
from ttlser import CustomTurtleSerializer
from xlsx2csv import Xlsx2csv, SheetNotFoundException
from pysercomb.pyr.types import ProtcurExpression, Quantity
from pyontutils.core import OntTerm as OTB, OntId as OIDB
from pyontutils.utils import isoformat
from pyontutils.namespaces import OntCuries, TEMP, sparc, NIFRID, definition
from pyontutils.namespaces import tech, asp, dim, unit, rdf, owl, rdfs
from sparcur import exceptions as exc
from sparcur.utils import log, logd  # FIXME fix other imports
from sparcur.utils import is_list_or_tuple


__type_registry = {None: None}
def register_type(cls, type_name):
    if type_name in __type_registry:
        raise ValueError(f'Cannot map {cls} to {type_name}. Type already present! '
                         f'{type_name} -> {__type_registry[type_name]}')

    __type_registry[type_name] = cls


def fromJson(blob):
    if isinstance(blob, dict):
        if 'type' in blob:
            t = blob['type']

            if t == 'identifier':
                type_name = blob['system']
            elif t in ('quantity', 'range'):
                type_name = t
            elif t not in __type_registry:
                breakpoint()
                raise NotImplementedError(f'TODO fromJson for type {t} '
                                          f'currently not implemented\n{blob}')
            else:
                type_name = t

            cls = __type_registry[type_name]
            if cls is not None:
                return cls.fromJson(blob)

        return {k:v
                if k == 'errors' or k.endswith('_errors') else
                fromJson(v)
                for k, v in blob.items()}

    elif isinstance(blob, list):
        return [fromJson(_) for _ in blob]
    else:
        return blob


xsd = rdflib.XSD
po = CustomTurtleSerializer.predicateOrder
po.extend((sparc.firstName,
           sparc.middleName,
           sparc.lastName,
           xsd.minInclusive,
           xsd.maxInclusive,
           TEMP.hasValue,
           TEMP.hasUnit,))

OntCuries({'orcid':'https://orcid.org/',
           'ORCID':'https://orcid.org/',
           'DOI':'https://doi.org/',
           'ror':'https://ror.org/',
           'dataset':'https://api.blackfynn.io/datasets/N:dataset:',
           'package':'https://api.blackfynn.io/packages/N:package:',
           'user':'https://api.blackfynn.io/users/N:user:',
           'bibo': 'http://purl.org/ontology/bibo/',  # crossref
           'prism.basic': 'http://prismstandard.org/namespaces/basic/2.1/',  # crossref
           'unit': str(unit),
           'dim': str(dim),
           'asp': str(asp),
           'tech': str(tech),
           'awards':str(TEMP['awards/']),
           'sparc':str(sparc),})


def curies_runtime(base):
    """ base is .e.g https://api.blackfynn.io/datasets/{dataset_id}/ """

    return {
        'local': base,
        'contributor': base + 'contributors/',
        'subject': base + 'subjects/',
        'sample': base + 'samples/',
    }


class OntId(OIDB):
    pass
    #def atag(self, **kwargs):
        #if 'curie' in kwargs:
            #kwargs.pop('curie')
        #return hfn.atag(self.iri, self.curie, **kwargs)


class OntTerm(OTB, OntId):
    _known_no_label = 'dataset', 'pio.private'

    #def atag(self, curie=False, **kwargs):
        #return hfn.atag(self.iri, self.curie if curie else self.label, **kwargs)  # TODO schema.org ...

    _logged = set()

    @classmethod
    def fromJson(cls, blob):
        assert blob['identifier_type'] == cls.__name__
        identifier = blob['id']
        if isinstance(identifier, cls):
            return identifier
        else:
            return cls(identifier, label=blob['label'])

    @classmethod
    def _already_logged(cls, thing):
        case = thing in cls._logged
        if not case:
            cls._logged.add(thing)

        return case

    def asType(self, _class):
        return _class(self.iri)

    def asUri(self, asType=None):
        return (self.iri
                if asType is None else
                asType(self.iri))

    def asDict(self):
        out = {
            'type': 'identifier',
            'system': self.__class__.__name__,
            #'id': self.iri,
            'id': self,  # XXX
            'label': self.label,
        }
        if hasattr(self, 'synonyms') and self.synonyms:
            out['synonyms'] = self.synonyms

        return out

    def asCell(self, sep='|'):
        if self.label is None:
            _id = self.curie if self.curie else self.iri
            if self.prefix not in self._known_no_label:
                if not self._already_logged(_id):
                    log.error(f'No label {_id}')

            return _id

        return self.label + sep + self.curie

    @property
    def triples_simple(self):
        # method name matches convention from neurondm
        # but I still don't really like this pattern
        # especially since this wouldn't be derived directly
        # from the json from but would/could hit the network again
        s = self.asUri(rdflib.URIRef)
        yield s, rdf.type, owl.Class
        if self.label:
            yield s, rdfs.label, rdflib.Literal(self.label)
        if self.definition:
            yield s, definition, rdflib.Literal(self.definition)
        if self.deprecated:
            yield s, owl.deprecated, rdflib.Literal(True)
        for syn in self.synonyms:
            s, NIFRID.synonyms, rdflib.Literal(syn)


class HasErrors:
    message_passing_key = 'keys_with_errors'
    _already_logged = set()

    def __init__(self, *args, pipeline_stage=None, **kwargs):
        try:
            super().__init__(*args, **kwargs)
        except TypeError as e:  # this is so dumb
            super().__init__()

        self._pipeline_stage = pipeline_stage
        self._errors_set = set()

    @staticmethod
    def errorInKey(data, key):
        mpk = HasErrors.message_passing_key
        if mpk in data:
            data[mpk].append(key)
        else:
            data[mpk] = [key]

    def addError(self, error, pipeline_stage=None, logfunc=None, blame=None, path=None):
        do_log = error not in self._already_logged
        if do_log:
            self._already_logged.add(error)

        stage = (pipeline_stage if pipeline_stage is not None
                 else (self._pipeline_stage if self._pipeline_stage
                       else self.__class__.__name__))
        b = len(self._errors_set)
        et = (error, stage, blame, path)
        self._last_error = et
        self._errors_set.add(et)
        a = len(self._errors_set)
        return a != b and do_log

    def _render(self, e, stage, blame, path):
        o = {'pipeline_stage': stage,
                'blame': blame,}  # FIXME

        if path is not None:
            o['file_path'] = path

        if isinstance(e, str):
            o['message'] = e
            o['type'] = None  # FIXME probably wan our own?

        elif isinstance(e, BaseException):
            o['message'] = str(e)
            o['type'] = str(type(e))

        else:
            raise TypeError(repr(e))

        log.debug(o)
        return o

    @property
    def _errors(self):
        for et in self._errors_set:
            yield self._render(*et)

    def embedLastError(self, data):
        self.embedErrors(data, el=[self._render(*self._last_error)])

    def embedErrors(self, data, el=tuple()):
        if not el:
            el = list(self._errors)

        if el:
            if 'errors' in data:
                data['errors'].extend(el)
            elif el:
                data['errors'] = el


def lj(j):
    """ use with log to format json """
    return '\n' + json.dumps(j, indent=2, cls=JEncode)


def dereference_all_identifiers(obj, stage, *args, path=None, addError=None, **kwargs):
    try:
        dict_literal = _json_identifier_expansion(obj)
    except idlib.exceptions.ResolutionError as e:
        if hasattr(obj, '_cooldown'):
            return obj._cooldown()  # trigger cooldown to simplify issues down the line

        oops = json_export_type_converter(obj)
        msg = (f'{stage.lifters.id} could not resolve '  # FIXME lifters sigh
               f'{type(obj)}: {oops} {obj.asUri()}')
        error = dict(error=msg,
                     pipeline_stage=stage.__class__.__name__,
                     blame='submission',
                     path=tuple(path))

        if addError:
            if addError(**error):
                logd.error(msg)
        else:
            return {'errors': [error]}


def _json_identifier_expansion(obj, *args, **kwargs):
    if isinstance(obj, oq.OntTerm):
        obj.__class__ = OntTerm  # that this works is amazing/terrifying
        return obj.asDict()

    elif isinstance(obj, idlib.Stream):
        if obj._id_class is str:
            return obj.identifier
        else:
            return obj.asDict()
    else:
        return obj


def json_identifier_expansion(obj, *args, path=None, **kwargs):
    """ expand identifiers to json literal form """
    try:
        return _json_identifier_expansion(obj, *args, **kwargs)
    except idlib.exceptions.ResolutionError as e:
        oops = json_export_type_converter(obj)
        msg = f'could not resolve {type(obj)}: {oops}'
        out = {'id': obj,
               'type': 'identifier',
               'system': obj.__class__.__name__,
               'errors': [{'message': msg, 'path': path}]}
        return out


def json_export_type_converter(obj):
    if isinstance(obj, deque):
        return list(obj)
    elif isinstance(obj, ProtcurExpression):
        return obj.json()
    elif isinstance(obj, PurePath):
        return obj.as_posix()
    elif isinstance(obj, Quantity):
        return obj.json()
    elif isinstance(obj, oq.OntTerm):
        return obj.iri
        #return obj.asDict()  # FIXME need a no network/scigraph version
    elif isinstance(obj, idlib.Stream) and hasattr(obj, '_id_class'):
        if obj._id_class is str:
            return obj.identifier
        else:
            return json_export_type_converter(obj.identifier)
            #return obj.asDict()  # FIXME need a no network/scigraph version
    elif isinstance(obj, datetime):
        return isoformat(obj)
    elif isinstance(obj, BaseException):
        # FIXME hunt down where these are sneeking in from
        return repr(obj)


class JEncode(json.JSONEncoder):

    def default(self, obj):
        new_obj = json_export_type_converter(obj)
        if new_obj is not None:
            return new_obj
        #else:
            #log.critical(f'{type(obj)} -> {obj}')
        #if isinstance(obj, ValueError):
            #breakpoint()

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def JFixKeys(obj):
    def jetc(o):
        out = json_export_type_converter(o)
        return out if out else o

    if isinstance(obj, dict):
        return {jetc(k): JFixKeys(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [JFixKeys(v) for v in obj]
    else:
        return obj


def JApplyRecursive(function, obj, *args,
                    condense=False,
                    skip_keys=tuple(),
                    preserve_keys=tuple(),
                    path=None,
                    **kwargs):
    """ *args, **kwargs, and path= are passed to the function """
    def testx(v):
        return (v is not None and
                not (not v and
                     (is_list_or_tuple(v) or isinstance(v, dict))))

    if path is None:
        path = []

    if isinstance(obj, dict):
        out = {k: JApplyRecursive(function, v, *args,
                                  condense=condense,
                                  skip_keys=skip_keys,
                                  preserve_keys=preserve_keys,
                                  path=path + [k],
                                  **kwargs)
               if k not in preserve_keys else v
               for k, v in obj.items() if k not in skip_keys}

        if condense:
            out = {k:v for k, v in out.items() if testx(v)}

        return function(out, *args, path=path, **kwargs)

    elif is_list_or_tuple(obj):
        out = [JApplyRecursive(function, v, *args,
                               condense=condense,
                               skip_keys=skip_keys,
                               preserve_keys=preserve_keys,
                               path=path + [i],
                               **kwargs)
               for i, v in enumerate(obj)]
        if condense:
            out = [v for v in out if testx(v)]

        return function(out, *args, path=path, **kwargs)

    else:
        return function(obj, *args, path=path, **kwargs)


def zipeq(*iterables):
    """ zip or fail if lengths do not match """

    sentinel = object()
    try:
        gen = itertools.zip_longest(*iterables, fillvalue=sentinel)
    except TypeError as e:
        raise TypeError(f'One of these is not iterable {iterables}') from e

    for zipped in gen:
        if sentinel in zipped:
            raise exc.LengthMismatchError('Lengths do not match! '
                                          'Did you remember to box your function?\n'
                                          f'{iterables}')

        yield zipped


class BlackfynnId(str):
    """ put all static information derivable from a blackfynn id here """
    def __new__(cls, id):
        # TODO validate structure
        self = super().__new__(cls, id)
        gotem = False
        for type_ in ('package', 'collection', 'dataset', 'organization'):
            name = 'is_' + type_
            if not gotem:
                gotem = self.startswith(f'N:{type_}:')
                setattr(self, name, gotem)
            else:
                setattr(self, name, False)

        return self

    @property
    def uri_api(self):
        # NOTE: this cannot handle file ids
        if self.is_dataset:
            endpoint = 'datasets/' + self.id
        elif self.is_organization:
            endpoint = 'organizations/' + self.id
        else:
            endpoint = 'packages/' + self.id

        return 'https://api.blackfynn.io/' + endpoint

    def uri_human(self, prefix):
        # a prefix is required to construct these
        return self  # TODO


class BlackfynnInst(BlackfynnId):
    # This isn't equivalent to BlackfynnRemote
    # because it needs to be able to obtain the
    # post pipeline data about that identifier
    @property
    def uri_human(self):
        pass


class JTList:
    pass


class JTDict:
    pass


def JT(blob):
    """ this is not a class but is a function hacked to work like one """
    def _populate(blob, top=False):
        if isinstance(blob, list) or isinstance(blob, tuple):
            # TODO alternatively if the schema is uniform, could use bc here ...
            def _all(self, l=blob):  # FIXME don't autocomplete?
                keys = set(k for b in l
                           if isinstance(b, dict)
                           for k in b)
                obj = {k:[] for k in keys}
                _list = []
                _other = []
                for b in l:
                    if isinstance(b, dict):
                        for k in keys:
                            if k in b:
                                obj[k].append(b[k])
                            else:
                                obj[k].append(None)

                    elif any(isinstance(b, t) for t in (list, tuple)):
                        _list.append(JT(b))

                    else:
                        _other.append(b)
                        for k in keys:
                            obj[k].append(None)  # super inefficient

                if _list:
                    obj['_list'] = JT(_list)

                if obj:
                    j = JT(obj)
                else:
                    j = JT(blob)

                if _other:
                    #obj['_'] = _other  # infinite, though lazy
                    setattr(j, '_', _other)

                setattr(j, '_b', blob)
                #lb = len(blob)
                #setattr(j, '__len__', lambda: lb)  # FIXME len()
                return j

            def it(self, l=blob):
                for b in l:
                    if any(isinstance(b, t) for t in (dict, list, tuple)):
                        yield JT(b)
                    else:
                        yield b

            if top:
                # FIXME iter is non homogenous
                return [('__iter__', it), ('_all', property(_all))]
            #elif not [e for e in b if isinstance(self, dict)]:
                #return property(id)
            else:
                # FIXME this can render as {} if there are no keys
                return property(_all)
                #obj = {'_all': property(_all),
                       #'_l': property(it),}

                #j = JT(obj)
                #return j

                #nl = JT(obj)
                #nl._list = blob
                #return property(it)

        elif isinstance(blob, dict):
            if top:
                out = [('_keys', tuple(blob))]
                for k, v in blob.items():  # FIXME normalize keys ...
                    nv = _populate(v)
                    out.append((k, nv))
                    #setattr(cls, k, nv)
                return out
            else:
                return JT(blob)

        else:
            if top:
                raise exc.UnhandledTypeError('asdf')
            else:
                @property
                def prop(self, v=blob):
                    return v

                return prop

    def _repr(self, b=blob):  # because why not
        return 'JT(\n' + lj(b) + '\n)'

    def query(self, *path):
        """ returns None at first failure """
        j = self
        for key in path:
            j = getattr(j, key, None)
            if j is None:
                return

        return j

    # additional thought required for how to integrate these into this
    # shambling abomination
    #adops
    #dt = DictTransformer

    #cd = {k:v for k, v in _populate(blob, True)}

    # populate the top level
    cd = {k:v for k, v in ((a, b) for t in _populate(blob, True)
                           for a, b in (t if isinstance(t, list) else (t,)))}
    cd['__repr__'] = _repr
    cd['query'] = query

    if isinstance(blob, dict):
        type_ = JTDict
    elif isinstance(blob, list):
        type_ = JTList
    else:
        type_ = object

    nc = type('JT' + str(type(blob)), (type_,), cd)  # use object to prevent polution of ns
    #nc = type('JT' + str(type(blob)), (type(blob),), cd)
    return nc()


class AtomicDictOperations:
    """ functions that modify dicts in place """

    # note: no delete is implemented at the moment ...
    # in place modifications means that delete can loose data ...

    __empty_node_key = object()

    @staticmethod
    def apply(function, *args,
              source_key_optional=False,
              extra_error_types=tuple(),
              failure_value=None):
        error_types = (exc.NoSourcePathError,) + extra_error_types
        try:
            return function(*args)
        except error_types as e:
            if not source_key_optional:
                raise e
            else:
                logd.debug(e)
                return failure_value
        except exc.LengthMismatchError as e:
            raise e

    @staticmethod
    def add(data, target_path, value, fail_on_exists=True, update=False):
        # type errors can occur here ...
        # e.g. you try to go to a string
        if not [_ for _ in (list, tuple) if isinstance(target_path, _)]:
            msg = f'target_path is not a list or tuple! {type(target_path)}'
            raise TypeError(msg)

        target_prefixes = target_path[:-1]
        target_key = target_path[-1]
        target = data
        for i, target_name in enumerate(target_prefixes):
            if target_name is int:  # add same value to all objects in list
                if not is_list_or_tuple(target):
                    msg = (f'attempt to add to all elements of not a list '
                           f'{type(target)} target_path was {target_path} '
                           f'target_name was {target_name}')
                    raise TypeError(msg)
                # LOL PYTHON namespaces
                [AtomicDictOperations.add(subtarget, target_path[i + 1:], value)
                 for subtarget in target]
                return  # int terminates this level of an add

            if target_name not in target:  # TODO list indicies XXX that is really append though ...
                target[target_name] = {}

            target = target[target_name]

        if update:
            pass
        elif fail_on_exists and target_key in target:
            raise exc.TargetPathExistsError(f'A value already exists at path {target_path}\n'
                                            f'{lj(data)}')

        target[target_key] = value

    @classmethod
    def update(cls, data, target_path, value):
        cls.add(data, target_path, value, update=True)

    @classmethod
    def get(cls, data, source_path, on_failure=None):
        """ get stops at lists because the number of possible issues explodes
            and we don't hand those here, if you encounter that, use this
            primitive to get the list, then use it again on the members in
            the function making the call where you have the information needed
            to figure out how to handle the error """

        try:
            source_key, node_key, source = cls._get_source(data, source_path)
            return source[source_key]
        except BaseException as e:
            if on_failure is not None:
                return on_failure
            else:
                raise e

    @classmethod
    def pop(cls, data, source_path):
        """ allows us to document removals """
        source_key, node_key, source = cls._get_source(data, source_path)
        if isinstance(source, tuple):
            value = source[source_key]
            new_node = tuple(v for i, v in enumerate(source) if i != source_key)
            parent_source_key, parent_node_key, parent_source = cls._get_source(data, source_path[:-1])
            assert node_key == parent_source_key
            # FIXME will fail if parent_source is a tuple
            parent_source[node_key] = new_node
            return value
        else:
            return source.pop(source_key)

    @classmethod
    def copy(cls, data, source_path, target_path):
        cls._copy_or_move(data, source_path, target_path)

    @classmethod
    def move(cls, data, source_path, target_path):
        cls._copy_or_move(data, source_path, target_path, move=True)

    @staticmethod
    def _get_source(data, source_path):
        #print(source_path, target_path)
        source_prefixes = source_path[:-1]
        source_key = source_path[-1]
        yield source_key  # yield this because we don't know if move or copy
        source = data
        for node_key in source_prefixes:
            if isinstance(source, dict):
                if node_key in source:
                    source = source[node_key]
                else:
                    # don't move if no source
                    msg = f'did not find {node_key!r} in {tuple(source.keys())}'
                    raise exc.NoSourcePathError(msg)
            elif is_list_or_tuple(source):
                if not isinstance(node_key, int):
                    raise TypeError(f'Wrong type for node_key {type(node_key)} {node_key}. '
                                    'Expected int.')
                source = source[node_key]  # good good let the index errors flow through you
            else:
                raise TypeError(f'Unsupported type {type(source)} for {lj(source)}')

        # for move
        yield (node_key if source_prefixes else
               AtomicDictOperations.__empty_node_key)

        if isinstance(source, dict):
            if source_key not in source:
                try:
                    msg = f'did not find {source_key!r} in {tuple(source.keys())}'
                    raise exc.NoSourcePathError(msg)
                except AttributeError as e:
                    raise TypeError(f'value at {source_path} has wrong type!{lj(source)}') from e
                    #log.debug(f'{source_path}')

        elif is_list_or_tuple(source):
            try:
                source[source_key]
            except IndexError as e:
                msg = f'There is not a {source_key}th value in {lj(source)}'
                raise exc.NoSourcePathError(msg) from e

        else:
            raise TypeError(f'Unsupported type {type(source)} for {lj(source)}')

        yield source

    @classmethod
    def _copy_or_move(cls, data, source_path, target_path, move=False):
        """ if exists ... """
        source_key, node_key, source = cls._get_source(data, source_path)
        # do not catch errors here, deal with that in the callers that people use directly
        if move:
            _parent = source  # incase something goes wrong
            source = source.pop(source_key)
        else:
            source = source[source_key]

            if source != data:  # this should .. always happen ???
                source = copy.deepcopy(source)  # FIXME this will mangle types e.g. OntId -> URIRef
                # copy first then modify means we need to deepcopy these
                # otherwise we would delete original forms that were being
                # saved elsewhere in the schema for later
            else:
                raise BaseException('should not happen?')

        try:
            cls.add(data, target_path, source)
        finally:
            # this will change key ordering but
            # that is expected, and if you were relying
            # on dict key ordering HAH
            if move and  node_key is not AtomicDictOperations.__empty_node_key:
                _parent[node_key] = source


adops = AtomicDictOperations()


class _DictTransformer:
    """ transformations from rules """

    @staticmethod
    def BOX(function):
        """ Combinator that takes a function and returns a version of
            that function whose return value is boxed as a tuple.
            This makes it _much_ easier to understand what is going on
            rather than trying to count commas when trying to count
            how many return values are needed for a derive function """

        @wraps(function)
        def boxed(*args, **kwargs):
            return function(*args, **kwargs),

        return boxed

    @staticmethod
    def add(data, adds):
        """ adds is a list (or tuples) with the following structure
            [[target-path, value], ...]
        """
        for target_path, value in adds:
            adops.add(data, target_path, value)

    @staticmethod
    def update(data, updates, source_key_optional=False):
        """ updates is a list (or tuples) with the following structure
            [[path, function], ...]
        """

        for path, function in updates:
            value = adops.get(data, path)
            new = function(value)
            adops.update(data, path, new)

    @staticmethod
    def get(data, gets, source_key_optional=False):
        """ gets is a list with the following structure
            [source-path ...] """

        for source_path in gets:
            yield adops.apply(adops.get, data, source_path,
                              source_key_optional=source_key_optional)

    @staticmethod
    def pop(data, pops, source_key_optional=False):
        """ pops is a list with the following structure
            [source-path ...] """

        for source_path in pops:
            yield adops.apply(adops.pop, data, source_path,
                              source_key_optional=source_key_optional)

    @staticmethod
    def delete(data, deletes, source_key_optional=False):
        """ delets is a list with the following structure
            [source-path ...]
            THIS IS SILENT YOU SHOULD USE pop instead!
            The tradeoff is that if you forget to express
            the pop generator it will also be silent until
            until schema catches it """

        for source_path in deletes:
            adops.pop(data, source_path)

    @staticmethod
    def copy(data, copies, source_key_optional=False):  # put this first to clarify functionality
        """ copies is a list wth the following structure
            [[source-path, target-path] ...]
        """
        for source_path, target_path in copies:
            # don't need a base case for thing?
            # can't lift a dict outside of itself
            # in this context
            adops.apply(adops.copy, data, source_path, target_path,
                        source_key_optional=source_key_optional)

    @staticmethod
    def move(data, moves, source_key_optional=False):
        """ moves is a list with the following structure
            [[source-path, target-path] ...]
        """
        for source_path, target_path in moves:
            adops.apply(adops.move, data, source_path, target_path,
                        source_key_optional=source_key_optional)

    @classmethod
    def derive(cls, data, derives, source_key_optional=True, empty='CULL', cheaty_face=None):
        """ [[[source-path, ...], function, [target-path, ...]], ...] """
        # if you have source key option True and empty='OK' you will get loads of junk
        allow_empty = empty == 'OK' and not empty == 'CULL'
        error_empty = empty == 'ERROR'
        def empty(value):
            empty = (value is None or
                     hasattr(value, '__iter__')
                     and not len(value))
            if empty and error_empty:
                raise ValueError(f'value to add may not be empty!')
            return empty or allow_empty and not empty

        failure_value = tuple()
        for source_paths, derive_function, target_paths in derives:
            # FIXME zipeq may cause adds to modify in place in error?
            # except that this is really a type checking thing on the function
            def defer_get(*get_args):
                """ if we fail to get args then we can't gurantee that
                    derive_function will work at all so we wrap the lot """
                args = cls.get(*get_args)
                return derive_function(*args)

            def express_zip(*zip_args):
                return tuple(zipeq(*zip_args))

            try:
                if not target_paths:
                    # allows nesting
                    adops.apply(defer_get, data, source_paths,
                                source_key_optional=source_key_optional)
                    continue

                cls.add(data,
                        ((tp, v) for tp, v in
                         adops.apply(express_zip, target_paths,
                                     adops.apply(defer_get, data, source_paths,
                                                 source_key_optional=source_key_optional),
                                     source_key_optional=source_key_optional,
                                     extra_error_types=(TypeError,),
                                     failure_value=tuple())
                        if not empty(v)))
            except TypeError as e:
                log.error('wat')
                idmsg = data['id'] if 'id' in data else ''
                raise TypeError(f'derive failed\n{source_paths}\n'
                                f'{derive_function}\n{target_paths}\n'
                                f'{idmsg}\n') from e

    @staticmethod
    def _derive(data, derives, source_key_optional=True, allow_empty=False):
        # OLD
        """ derives is a list with the following structure
            [[[source-path, ...], derive-function, [target-path, ...]], ...]

        """
        # TODO this is an implementaiton of copy that has semantics for handling lists
        for source_path, function, target_paths in derives:
            source_prefixes = source_path[:-1]
            source_key = source_path[-1]
            source = data
            failed = False
            for i, node_key in enumerate(source_prefixes):
                log.debug(lj(source))
                if node_key in source:
                    source = source[node_key]
                else:
                    msg = f'did not find {node_key} in {source.keys()}'
                    if not i:
                        log.error(msg)
                        failed = True
                        break
                    raise exc.NoSourcePathError(msg)
                if isinstance(source, list) or isinstance(source, tuple):
                    new_source_path = source_prefixes[i + 1:] + [source_key]
                    new_target_paths = [tp[i + 1:] for tp in target_paths]
                    new_derives = [(new_source_path, function, new_target_paths)]
                    for sub_source in source:
                        _DictTransformer.derive(sub_source, new_derives,
                                                source_key_optional=source_key_optional)

                    return  # no more to do here

            if failed:
                continue  # sometimes things are missing we continue to others

            if source_key not in source:
                msg = f'did not find {source_key} in {source.keys()}'
                if source_key_optional:
                    return logd.info(msg)
                else:
                    raise exc.NoSourcePathError(msg)

            source_value = source[source_key]

            new_values = function(source_value)
            if len(new_values) != len(target_paths):
                log.debug(f'{source_paths} {target_paths}')
                raise TypeError(f'wrong number of values returned for {function}\n'
                                f'was {len(new_values)} expect {len(target_paths)}')
            #temp = b'__temporary'
            #data[temp] = {}  # bytes ensure no collisions
            for target_path, value in zip(target_paths, new_values):
                if (not allow_empty and
                    (value is None or
                     hasattr(value, '__iter__') and not len(value))):
                    raise ValueError(f'value to add to {target_path} may not be empty!')
                adops.add(data, target_path, value, fail_on_exists=True)
                #heh = str(target_path)
                #data[temp][heh] = value
                #source_path = temp, heh  # hah
                #self.move(data, source_path, target_path)

                #data.pop(temp)


    @classmethod
    def subpipeline(cls, data, runtime_context, subpipelines, update=True,
                    source_key_optional=True, lifters=None):
        """
            [[[[get-path, add-path], ...], pipeline-class, target-path], ...]

            NOTE: this function is a generator, you have to express it!
        """

        class DataWrapper:
            def __init__(self, data):
                self.data = data

        prepared = []
        for get_adds, pipeline_class, target_path in subpipelines:
            selected_data = {}
            ok = True
            for get_path, add_path in get_adds:
                try:
                    value = adops.get(data, get_path)
                    if add_path is not None:
                        adops.add(selected_data, add_path, value)
                    else:
                        selected_data = value
                except exc.NoSourcePathError as e:
                    if source_key_optional:
                        yield get_path, e, pipeline_class
                        ok = False
                        break  # breaks the inner loop
                    else:
                        raise e

            if not ok:
                continue

            log.debug(lj(selected_data))
            prepared.append((target_path, pipeline_class, DataWrapper(selected_data),
                             lifters, runtime_context))

        function = adops.update if update else adops.add
        for target_path, pc, *args in prepared:
            p = pc(*args)
            if target_path is not None:
                try:
                    function(data, target_path, p.data)
                except BaseException as e:
                    import inspect
                    if isinstance(pc, object):
                        pi, pc = pc, pc.__class__

                    __file = inspect.getsourcefile(pc)
                    __line = inspect.getsourcelines(pc)[-1]
                    if hasattr(p, 'path'):
                        __path = f'"{p.path}"'
                    else:
                        __path = 'unknown input'

                    raise exc.SubPipelineError(
                        f'Error while processing {p}.data for\n{__path}\n'
                        f'{__file} line {__line}') from e

            else:
                p.data  # trigger the pipeline since it is stateful

            yield p

    @staticmethod
    def lift(data, lifts, source_key_optional=True):
        """ 
        lifts are lists with the following structure
        [[path, function], ...]

        the only difference from derives is that lift
        overwrites the underlying data (e.g. a filepath
        would be replaced by the contents of the file)

        """

        for path, function in lifts:
            try:
                old_value = adops.get(data, path)
            except exc.NoSourcePathError as e:
                if source_key_optional:
                    logd.exception(str(type(e)))
                    continue
                else:
                    raise e

            new_value = function(old_value)
            adops.add(data, path, new_value, fail_on_exists=False)

DictTransformer = _DictTransformer()


def copy_all(source_parent, target_parent, *fields):
    return [[source_parent + [field], target_parent + [field]]
            for field in fields]

def normalize_tabular_format(project_path):
    kwargs = {
        'delimiter' : '\t',
        'skip_empty_lines' : True,
        'outputencoding': 'utf-8',
    }
    sheetid = 0
    for xf in project_path.rglob('*.xlsx'):
        xlsx2csv = Xlsx2csv(xf, **kwargs)
        with open(xf.with_suffix('.tsv'), 'wt') as f:
            try:
                xlsx2csv.convert(f, sheetid)
            except SheetNotFoundException as e:
                log.warning(f'Sheet weirdness in {xf}\n{e}')


def extract_errors(thing, path=None):
    """ recursively extract errors """
    if path is None:
        path = []

    if isinstance(thing, dict):
        for k, v in thing.items():
            if k == 'errors':
                for error in v:
                    yield tuple(path), error
            else:
                yield from extract_errors(v, path + [k])

    elif isinstance(thing, list):
        for i, v in enumerate(thing):
            yield from extract_errors(v, path + [i])


def get_all_errors(_with_errors):
    """ A better and easier to interpret measure of completeness. """
    # TODO deduplicate by tracing causes
    # TODO if due to a missing required report expected value of missing steps
    return list(extract_errors(_with_errors))


def get_by_invariant_path(errors):
    dd = defaultdict(list)
    for e in errors:
        k = tuple(int if isinstance(element, int)
                  else element for element in e['path'])
        dd[k].append(e)

    by_invariant_path = dict(dd)
    return by_invariant_path


def make_path_error_report(by_invariant_path):
    # FIXME to obtain the full final path you have to know where
    # if anywhere the schema being validated is located in the output
    path_error_report = {JPointer.fromList(['-1' if e is int else str(e)
                                            # -1 to indicate all instead of *
                                            for e in k]):
                         {'error_count': len(v),
                          'messages': sorted(set(e['message'] for e in v))}
                         for k, v in by_invariant_path.items()}
    return path_error_report


def hashable_converter(v):
    try:
        hash(v)
        return v
    except TypeError as e:
        if isinstance(v, dict):
            return frozenset((k, hashable_converter(v)) for k, v in v.items())
        else:
            # HACK will fail on non iterables
            return tuple(hashable_converter(v) for v in v)


def condense_over_stage(errors):
    """ condense shadoweded errors into their caster """
    dd = defaultdict(list)
    for error in errors:
        #log.critical(error)
        new_error = {k:hashable_converter(v) for k, v in error.items()}
        copy.deepcopy(error)
        new_error.pop('pipeline_stage', None)
        if 'schema_path' in new_error:  # XXX non json schema errors
            sp = new_error['schema_path']
            if 'inputs' in sp:
                # account for the fact that inputs are checked twice right now
                index = sp.index('inputs')
                new_error['schema_path'] = sp[index + 3:]
                # FIXME somehow the path and schema_path get misaligned

        dd[hashable_converter(new_error.items())].append(error)

    merged = dict(dd)
    compacted = []
    for frozen_new_error, these_errors in merged.items():
        if len(these_errors) > 1:
            stages = [e['pipeline_stage'] for e in these_errors]
            error = copy.deepcopy(these_errors[0])
            error['pipeline_stage'] = stages
            error['total_stage_errors'] = len(these_errors)
            compacted.append(error)
        else:
            compacted.extend(these_errors)

    return compacted, merged  # for maximum confusion


def merge_error_paths(path_errors):
    for path, error in path_errors:
        #log.critical((path, error))
        error = copy.deepcopy(error)
        oldpath = list(error['path']) if 'path' in error else []  # XXX non json schema errors
        error['path'] = list(path) + oldpath
        yield error


def compact_errors(path_errors):
    """ compact repeated errors in lists by lifting the list index
        to the int type, making it constaint over all indexes and
        making it possible to identify paths with the same structure """

    errors = list(merge_error_paths(path_errors))
    condensed, merged = condense_over_stage(errors)
    by_invariant_path = get_by_invariant_path(condensed)
    path_error_report = make_path_error_report(by_invariant_path)

    compacted = []
    for ipath, these_errors in by_invariant_path.items():
        if int in ipath:
            index = ipath.index(int)
            ints = sorted(set(e['path'][index] for e in these_errors))
            error = copy.deepcopy(these_errors[0])
            error['path'][index] = ints
            error['total_errors'] = len(these_errors)
            compacted.append(error)
        else:
            compacted.extend(these_errors)

    return compacted, path_error_report, by_invariant_path, errors


class JPointer(str):
    """ a class to mark json pointers for resolution """

    @staticmethod
    def pathFromSchema(schema_path):
        gen = (e for e in schema_path)
        while True:
            try:
                element = next(gen)
                if element in ('oneOf', 'anyOf', 'allOf'):
                    next(gen)
                    continue

                if element == 'properties':
                    yield next(gen)
                elif element == 'items':
                    # NOTE it is ok to use int in cases like this because even
                    # though it is in principle ambiguous with a case where
                    # someone happens to be using the python function int
                    # as a key in a dict they would never be able to serialize
                    # that to json without some major conversions, so I am
                    # ruling that it is safe to lift to type here since this
                    # is JPointer not rando dict pointer
                    yield int

            except StopIteration:
                break

    @classmethod
    def fromList(cls, iterable):
        return cls('#/' + '/'.join(iterable))
        
    def asList(self):
        return self.split('/')


# register idlib classes for fromJson
[register_type(i, i.__name__) for i in
 (idlib.Ror, idlib.Doi, idlib.Orcid, idlib.Pio, idlib.Rrid, OntTerm)]



import unittest
from pathlib import Path
import pytest
from sparcur import apinat
from sparcur import pipelines as pipes
from .common import examples_root


class TestApiNAT(unittest.TestCase):

    source = Path(__file__).parent / 'apinatomy'

    def test_load(self):
        m = pipes.ApiNATOMY(Path(self.source, 'keast-spinal-map.json'))
        # FIXME I think only the model conforms to the schema ?
        #g = pipes.ApiNATOMY(Path(self.source, 'apinatomy-generated.json'))
        #rm = pipes.ApiNATOMY(Path(self.source, 'apinatomy-resourceMap.json'))
        m.data.keys()
        assert 'resources' in m.data, m.data

    def test_export(self):
        m = pipes.ApiNATOMY(self.source / 'keast-spinal-map.json')
        # FIXME need a way to combine this that doesn't require
        # the user to know how to compose these, just send a message
        # to one of them they should be able to build the other from
        # the information at hand
        r = pipes.ApiNATOMY_rdf(m)
        graph = r.data
        assert list(graph[:apinat.readable.name:]), graph.ttl


class TestDatasetDescription(unittest.TestCase):
    source = examples_root / 'dd-pie.csv'

    def test_dd_pie_p(self):
        p = pipes.DatasetDescriptionFilePipeline(self.source, None, None)
        data = p.data
        # TODO test as subpipeline ?



import pathlib
#from xml.etree import ElementTree as etree  # pypy3 2m18.445s  # 3.7 1m56.417s  # WAT
from lxml import etree  # python3.7 0m38.388s
from pyontutils.utils import subclasses
from .. import schemas as sc
from ..core import HasErrors, register_type
from ..utils import log, logd

if etree.__name__ == 'lxml.etree':
    parser = etree.XMLParser(
        collect_ids=False,
        #remove_blank_text=True,
    )


class ExtractXml:
    def __new__(cls, path):
        inst = None
        for c in cls.classes:
            #log.debug(c)
            if inst is not None:
                _inst = inst
                inst = c.fromExisting(inst)
                if inst._isXml:
                    delattr(_inst, 'e')  # prevent __del__ from zapping e
                    del _inst
            else:
                inst = c(path)
                if not inst._isXml:
                    inst = NotXml(path)
                    msg = f'path is not a valid xml file! {path}'
                    if inst.addError(msg, blame='submission', path=path):
                        logd.error(msg)

                    return inst

            if inst.typeMatches():
                return inst
            #else:
                #if hasattr(inst, 'e'):
                    #log.debug(inst.e.getroot().tag)

        else:
            path.xopen()
            breakpoint()
            raise TypeError(f'FIXME converter not implemented for whatever this is {path}')


class NotXml(HasErrors):

    _isXml = False
    mimetype = None

    def __init__(self, path):
        super().__init__()
        self.path = path

    def asDict(self):
        out = {}
        self.embedErrors(out)
        return out

    def typeMatches(self):
        return False


class XmlSource(HasErrors):

    top_tag = object()  # should never be able to == anything
    mimetype = 'application/xml'

    @classmethod
    def fromExisting(cls, existing):
        self = object.__new__(cls)
        self.__dict__ = {k:v for k, v in existing.__dict__.items()}
        return self

    def __init__(self, path):
        super().__init__()
        self.path = path
        if etree.__name__ == 'lxml.etree':
            # FIXME this version is much faster but on a completely run of the
            # mill workload uses 4 gigs of memory vs 360mb for pure python
            # possibly relevant
            # https://benbernardblog.com/tracking-down-a-freaky-python-memory-leak-part-2/
            try:
                self.e = etree.parse(self.path.as_posix(), parser=parser)
                self._isXml = True
            except etree.XMLSyntaxError as e:
                self._isXml = False
                logd.exception(e)
                logd.error(f'error parsing {self.path}')
                return

            self.namespaces = {k if k else '_':v
                               for k, v in next(self.e.iter()).nsmap.items()}
        else:
            try:
                self.e = etree.parse(self.path)
                self._isXml = True
            except etree.ParseError as e:
                self._isXml = False
                logd.exception(e)
                logd.error(f'error parsing {self.path}')
                return

            self.namespaces = dict([node for (_, node) in
                                    etree.iterparse(self.path,
                                                    events=['start-ns'])])
            if '' in self.namespaces:
                self.namespaces['_'] = self.namespaces.pop('')

        self.xpath = self.mkx(self.e)

    if etree.__name__ == 'lxml.etree':
        def __del__(self):
            # slows things down a bit but saves a ton of
            # memory when parsing multiple files
            if hasattr(self, 'e'):
                self.e.getroot().clear()
                del self.e

    def isXml(self):
        return self._isXml

    def typeMatches(self):
        return self._isXml and self.e.getroot().tag == self.top_tag

    def mkx(self, element):
        if hasattr(element, 'xpath'):
            def xpath(path, e=element):
                return e.xpath(path, namespaces=self.namespaces)
        else:
            def xpath(path, e=element):
                if path.startswith('@'):
                    attribute = path.strip('@')
                    return e.get(attribute)

                elif '/@' in path:
                    path, attribute = path.split('/@')
                    return [m.get(attribute)
                            for m in e.findall(path, namespaces=self.namespaces)]

                elif path.endswith('/text()'):
                    path, _ = path.rsplit('/', 1)
                    return [m.text
                            for m in e.findall(path, namespaces=self.namespaces)]

                else:
                    return e.findall(path, namespaces=self.namespaces)

        return xpath

    def asDict(self, *args, **kwargs):  # FIXME data vs asDict
        data_in = self._condense(*args, **kwargs)
        self.embedErrors(data_in)
        return data_in

    def _condense(self, *args, **kwargs):
        et = tuple()
        if hasattr(self.asDict, 'schema'):
            _schema = self.asDict.schema.schema
        else:
            _schema = et

        # FIXME how to deal with combination bits in schema, ignore for now
        def condense(thing, schema=_schema):

            #print(thing, schema)
            if isinstance(thing, dict):
                nschema = schema['properties'] if 'properties' in schema else et
                return {k:nv
                        for k, v in thing.items()
                        for nv in (condense(v, nschema[k] if k in nschema else et),)
                        # stick with the pattern to filter null fields
                        # we probably don't need to work in reverse
                        # for this format
                        if nv is not None}

            elif isinstance(thing, list):
                nschema = schema['items'] if 'items' in schema else et
                if not thing:
                    if sc.not_array(schema):
                        return None
                    else:
                        return thing
                else:
                    condensed = [condense(v, nschema) for v in thing]
                    if len(condensed) == 1 and sc.not_array(schema):
                        return condensed[0]
                    else:
                        # by default if we lack type information don't unpack
                        # i.e., don't destroy anything unless we are told it is ok
                        return condensed
            else:
                return thing

        data_in = self._extract(*args, **kwargs)
        data_out = condense(data_in)
        for k, v in tuple(data_out.items()):
            if not v:
                data_out.pop(k)

        return data_out

    def _extract(self, *args, **kwargs):
        raise NotImplementedError('implement in subclasses')


hasSchema = sc.HasSchema()
@hasSchema.mark
class ExtractMBF(XmlSource):

    top_tag = '{http://www.mbfbioscience.com}mbf'
    mimetype = 'application/vnd.mbfbioscience.metadata+xml'

    @hasSchema.f(sc.MbfTracingSchema)
    def asDict(self, unique=True, guid=False):
        return self._condense(unique=unique, guid=guid)

    def _condense(self, unique=True, guid=False):
        data_in = super()._condense(unique=unique, guid=guid)
        if 'contours' in data_in:
            if not guid:
                # remove guid fields if present
                # option to retain if a use case for this appears
                for c in data_in['contours']:
                    c.pop('guid', None)

            if unique:
                # by default only store uniquely identified contours,
                # they may be identified by their list index but we
                # don't need that for this part of the metadata
                data_in['contours'] = [dict(s) for s in
                                       set(frozenset(d.items())
                                           for d in data_in['contours'])]

        return data_in

    def _extract(self, *args, **kwargs):
        subject = {
            'id':      self.xpath('_:sparcdata/_:subject/@subjectid'),
            'species': self.xpath('_:sparcdata/_:subject/@species'),
            'sex':     self.xpath('_:sparcdata/_:subject/@sex'),
            'age':     self.xpath('_:sparcdata/_:subject/@age'), # age at death ? (maximum age heh)
        }

        atlas = {
            'organ':        self.xpath('_:sparcdata/_:atlas/@organ'),
            'atlas_label':  self.xpath('_:sparcdata/_:atlas/@label'),
            'atlas_rootid': self.xpath('_:sparcdata/_:atlas/@rootid'),
        }

        images = [
            {'path_mbf': [pathlib.PureWindowsPath(p) for p in image_xpath('_:filename/text()')],
             'channels': [{'id':     channel_xpath('@id'),
                           'source': channel_xpath('@source'),}
                          for channel in image_xpath('_:channels/_:channel')
                          for channel_xpath in (self.mkx(channel),)],
            }
            for image in self.xpath('_:images/_:image')
            for image_xpath in (self.mkx(image),)
        ]

        # these are probably study targets ? or rois ? or what ...
        contours = [
            {'name':        xpath('@name'),
             'guid':        xpath('_:property[@name="GUID"]/_:s/text()'),
             'id_ontology': xpath('_:property[@name="TraceAssociation"]/_:s/text()'),
             # closed=true ??
            }
            for contour in self.xpath('_:contour')
            for xpath in (self.mkx(contour),)]

        return {
            'subject':  subject,
            'atlas':    atlas,
            'images':   images,
            'contours': contours,
        }


hasSchema = sc.HasSchema()
@hasSchema.mark
class ExtractNeurolucida(ExtractMBF):

    top_tag = '{http://www.mbfbioscience.com/2007/neurolucida}mbf'
    mimetype = 'application/vnd.mbfbioscience.neurolucida+xml'

    @hasSchema.f(sc.NeurolucidaSchema)
    def asDict(self, unique=True, guid=False):
        # pretty sure we can't use super().asDict() here because
        # the schemas will fight with eachother
        data_in = self._condense(unique=unique, guid=guid)
        self.embedErrors(data_in)
        return data_in


class ExtractZen(XmlSource):

    top_tag = 'CellCounter_Marker_File'
    mimetype = 'application/vnd.unknown.zen+xml'  # TODO

    def _extract(self):
        images = [{'path_zen':
                   [pathlib.PurePath(p) for p in
                    self.xpath('Image_Properties/Image_Filename/text()')],}]

        return {'images': images}


class ExtractLAS(XmlSource):
    """ Leica Application Suite format,
        probably internal given lack of xmlns """

    top_tag = 'Data'  # very helpful thanks guys >_<
    mimetype = 'application/vnd.leica.las+xml'  # TODO

    def typeMatches(self):
        return (super().typeMatches() and
                'LAS AF' in self.xpath('/Data/Image/Attachment/@Application'))

    def _extract(self):
        images = [{'path_las':
                   [pathlib.PureWindowsPath(p) for p in
                    self.xpath('ImageDescription/FileLocation/text()')],
                   'channels': []  # TODO
        }]

        return {'images': images}


class ExtractPVScan(XmlSource):
    """ Format for proprietary Prairie View Software made by Bruker Corporation
        Version matters for parsing.
        See https://www.openmicroscopy.org/community/viewtopic.php?f=13&t=7534
        Google search results are garbage when searching from the tags and
        for Prairie View without additional qualifiers specifically software or
        microscopy."""

    top_tag = 'PVScan'  # probably not sufficient ...
    mimetype = 'application/vnd.bruker.pvscan+xml'  # TODO

    def _extract(self):
        # TODO nothing of obvious relevance for metadata
        return {}

class ExtractPVVrecSE(XmlSource):
    """ An auxillary Prairie View file that is referenced from the main PVSync
        record in the VoltageRecording section """

    top_tag = 'VRecSessionEntry'
    mimetype = 'application/vnd.bruker.vrecsessionentry+xml'  # TODO

    def _extract(self):
        # TODO nothing of obvious relevance for metadata
        return {}


class ExtractPVExperiment(XmlSource):
    """ An auxillary Prairie View file that is referenced from the main PVSync
        record in the VoltageOutput section """

    # TODO FIXME gonna need some additional context from somewhere
    # xml files like these (missing xmlns) should flag a warning for
    # bad (solipsistic) data management practices

    top_tag = 'Experiment'  # very helpful xml schema creator guy >_<
    mimetype = 'application/vnd.bruker.experiment+xml'  # TODO

    def _extract(self):
        # TODO nothing of obvious relevance for metadata
        return {}


class ExtractZISRAWSUBBLOCK_METADATA(XmlSource):
    """ This is xml extracted from xml embedded in a
        ZISRAWSUBBLOCK section of a Zeiss czi file """

    top_tag = 'METADATA'  # LOL OH WOW >_< this just keeps getting better
    
    mimetype = 'application/vnd.zeiss.czi.ZISRAWSUBBLOCK.METADATA+xml'  # TODO

    def _extract(self):
        # TODO has time information and stage and focus
        # seems to be using file naming conventions to match
        # sidecar to tiff (urg)
        return {}


ExtractXml.classes = (*[c for c in subclasses(XmlSource)], XmlSource)

# FIXME not entirely clear that I am using type correctly here
[register_type(None, cls.mimetype) for cls in ExtractXml.classes
 # NotXml has no mimetype
 if cls.mimetype is not None]



from types import MappingProxyType
import rdflib
from ontquery.utils import mimicArgs
from pyontutils.core import OntGraph
from pyontutils.utils import Async, deferred
from pyontutils.namespaces import rdf, rdfs, owl, OntCuries
from pyontutils import combinators as cmb
import sparcur.schemas as sc
from sparcur.core import adops, OntId, OntTerm
from sparcur.utils import log, logd
from ttlser import CustomTurtleSerializer


elements = rdflib.Namespace('https://apinatomy.org/uris/elements/')
readable = rdflib.Namespace('https://apinatomy.org/uris/readable/')
ordered = rdflib.Namespace('https://apinatomy.org/uris/readable/ordered/')

# add apinatomy:Graph to ttlser topClasses
tc = CustomTurtleSerializer.topClasses
if readable.Graph not in tc:
    sec = CustomTurtleSerializer.SECTIONS
    CustomTurtleSerializer.topClasses = [readable.Graph] + tc
    CustomTurtleSerializer.SECTIONS = ('',) + sec

# add apinatomy:Graph as a header section marker
OntGraph.metadata_type_markers.append(readable.Graph)


OntCuries({'apinatomy': str(readable),
           'elements': str(elements),  # FIXME guranteed name collisions ...
           'ordered': str(ordered),
           'PMID': 'https://www.ncbi.nlm.nih.gov/pubmed/',
           # also just read this from the embedded local conventions
})


class NoIdError(Exception):
    """ blob has no id """


def make_classes(schema):
    types = {}

    def ref_to_list(ref):
        _jpath = ref.split('/')
        if _jpath[0] != '#':
            raise ValueError(ref)
        else:
            jpath = _jpath[1:]

        return jpath

    def deref(ref):
        return adops.get(schema, ref_to_list(ref))

    def allOf(obj):
        for o in obj['allOf']:
            if '$ref' in o:
                ref = o['$ref']
                if ref in types:
                    yield types[ref]
                else:
                    jpath = ref_to_list(ref)
                    no = adops.get(schema, jpath)
                    yield top(jpath[-1], no)
            else:
                log.debug(f'{obj}')

    def properties(obj):
        props = obj['properties']
        out = {}
        for name, vobj in props.items():
            @property
            def f(self, n=name):
                return self.blob[n]

            out[name] = f

        return out

    def top(cname, obj):
        deref, allOf, ref_to_list, schema, types  # python is dumb
        type_ = None
        if 'type' in obj:
            type_ = obj['type']
            parents = (Base,)

        elif 'allOf' in v:
            parents = tuple(allOf(obj))
            for c in parents:
                if hasattr(c, 'type'):
                    type_ = c.type

        if type_ is None:
            raise TypeError('wat')

        cd = {'type': type_,
              '_schema': obj,
        }

        if type_ == 'object':
            props = properties(obj)
            for n, f in props.items():
                cd[n] = f

        return type(cname, parents, cd)

    cs = []
    d = schema['definitions']
    for k, v in d.items():
        if not v:
            log.warning(f'empty definition for {k}')
            continue

        c = top(k, v)
        ref = f'#/definitions/{k}'
        types[ref] = c
        cs.append(c)

    return cs


class Base:

    @classmethod
    def fromRdf(cls, uri, graph, context=None):
        _, id = uri.rsplit('/', 1)
        blob = {'id': id}
        for p, o in graph[uri]:
            if p == rdf.type:
                if o != owl.NamedIndividual:
                    key = 'class'
                    _, value = o.rsplit('/', 1)
                else:
                    continue  # TODO s rdf:type apinatomy:External ??
            else:
                _, key = p.rsplit('/', 1)

                if isinstance(o, rdflib.Literal):
                    value = o.toPython()
                elif isinstance(o, rdflib.URIRef):
                    oid = OntId(o)
                    if oid.prefix == 'local':
                        value = oid.suffix
                    elif oid.prefix == 'apinatomy':  # FIXME hrm?
                        value = oid.suffix
                    else:
                        value = oid.curie  # FIXME external is tricky
                        log.warning(f'{oid!r}')
                elif isinstance(o, rdflib.BNode):
                    raise NotImplementedError(f'a bit more complex ...')
                else:
                    raise NotImplementedError(f'{o}')
                
            if key in cls.objects_ordered:  # ordered representation takes priority
                raise NotImplementedError('TODO this is quite a bit more complex')
                if key in blob:
                    blob[key].append(value)
                else:
                    blob[key] = [value]

            elif key in cls.objects_multi:
                if key in blob:
                    blob[key].append(value)
                else:
                    blob[key] = [value]

            else:
                blob[key] = value

        return cls(blob, context)

    def __init__(self, blob, context=None, label_suffix=''):
        self.blob = blob
        self.context = context
        self.label_suffix = label_suffix
        try:
            self.id = blob['id'].replace(' ', '-')  #FIXME
        except KeyError as e:
            raise NoIdError(f'id not in {blob}') from e
        except AttributeError:
            pass  # new impl uses properties to access the blob

        if context is not None:
            self.s = context[self.id]
        else:
            self.s = None

        try:
            self.name = blob['name'] if 'name' in blob else self.id
        except AttributeError:
            pass  # new impl uses properties to access the blob

        if 'class' in blob:
            assert self.__class__.__name__ == blob['class']

    @property
    def cname(self):
        return self.__class__.__name__


class Graph(Base):

    @classmethod
    def fromRdf(cls, graph):
        iri = graph.boundIdentifier
        context = rdflib.Namespace(iri + '/ids/')
        # TODO removing things from the trie is not implemented ...
        #d = OntCuries._dict
        #d.pop('local', None)
        #d['local'] = str(context)
        #OntCuries.reset()
        OntCuries({'local': str(context)})
        _, id = iri.rsplit('/', 1)

        blob = {'id': id}
        resources = {id: blob}
        map = {'id': id,
               'resources': resources}

        for p, o in graph[iri:]:
            if p == rdf.type:
                blob['class'] = o.toPython()

            else:
                _, key = p.rsplit('/', 1)
                blob[key] = o.toPython()

        for s in graph[:rdf.type:owl.NamedIndividual]:
            for element in graph[s:rdf.type]:
                if element != owl.NamedIndividual:
                    _, class_ = element.rsplit('/', 1)
                    resource = getattr(cls, class_).fromRdf(s, graph, context)
                    # FIXME we should really keep the internal representation
                    # around instead of just throwing it away
                    resources[resource.id] = resource.blob

        for s in graph[:rdf.type:owl.Class]:
            # FIXME s rdf:type elements:External ??
            resource = External.fromRdf(s, graph, context)
            resources[resource.id] = resource.blob

        return cls(map)

    def __init__(self, map):
        self.map = map
        self.resources = map['resources']  # FIXME resources.pop('waitingList') ??
        self.prefixes = {}  # TODO curie mapping
        self.id = self.map['id'].replace(' ', '-')  # FIXME

        self.blob = self.resources[self.map['id']]
        # FIXME should we require name and abbrev and make missing a fatal error?
        self.name = self.blob['name'] if 'name' in self.blob else None
        self.abbreviation = self.blob['abbreviation'] if 'abbreviation' in self.blob else None
        self.label_suffix = f' ({self.abbreviation})' if self.abbreviation else ''
        #apinscm.validate(self.blob)  # TODO

    @property
    def context(self):
        return rdflib.Namespace(f'{self.iri}/ids/')

    @property
    def triples(self):
        self.iri = rdflib.URIRef(f'https://apinatomy.org/uris/models/{self.id}')
        yield self.iri, rdf.type, readable.Graph
        yield self.iri, readable.name, rdflib.Literal(self.name)
        yield self.iri, readable.abbreviation, rdflib.Literal(self.abbreviation)
        externals = []
        for id, blob in self.resources.items():
            if 'class' not in blob:
                logd.warning(f'no class in\n{blob!r} for {id}')
                continue
            elif blob['class'] == 'Graph':
                continue

            obj = getattr(self, blob['class'])(blob, self.context, self.label_suffix)

            if blob['class'] == 'External':
                # defer lookup
                externals.append(obj)
                continue

            yield from obj.triples()

        Async()(deferred(lambda x: x._term)(e) for e in externals)
        for e in externals:
            yield from e.triples()

    triples_gen = triples  # match the converters interface

    def populate(self, graph):
        #[graph.add(t) for t in self.triples]
        for t in self.triples:
            s, p, o = t
            if s is None:
                raise BaseException(f'{t}')

            graph.add(t)

    def graph(self):
        g = OntGraph()
        OntCuries.populate(g)
        self.populate(g)
        g.bind('local', self.context)
        g.bind('apinatomy', readable)  # FIXME populate from store
        g.bind('elements', elements)
        return g


class BaseElement(Base):
    key = None
    annotations = tuple()
    generics = tuple()
    objects = tuple()
    objects_multi = tuple()
    objects_ordered = tuple()
    objects_ordered_succession = MappingProxyType({})

    def triples(self):
        yield from self.triples_resource()
        yield from self.triples_external()
        yield from self.triples_annotations()
        yield from self.triples_generics()
        yield from self.triples_objects()
        yield from self.triples_objects_multi()
        yield from self.triples_objects_ordered()

    def triples_resource(self):
        yield self.s, rdf.type, owl.NamedIndividual
        yield self.s, rdf.type, elements[f'{self.cname}']
        yield self.s, readable['name'], rdflib.Literal(self.name)
        yield self.s, rdfs.label, rdflib.Literal(self.name + self.label_suffix)

    def triples_annotations(self):
        for key in self.annotations:
            if key in self.blob:
                value = self.blob[key]
                yield self.s, readable[key], rdflib.Literal(value)

    def triples_generics(self):
        for key in self.generics:
            # FIXME has + key.capitalize()?
            if key in self.blob:
                value = self.blob[key]
                yield self.s, readable[key], readable[value]
            else:
                # TODO the logic for warnings is a bit more complex
                # e.g. when should a lyph have a topology? not entirely clear
                # is there a default topology?
                if key not in ('topology', 'conveyingType'):
                    log.warning(f'{key} not in {self.blob}')

    def triples_objects(self):
        for key in self.objects:
            if key in self.blob:
                value = self.blob[key]
                value = value.replace(' ', '-')  # FIXME require no spaces in internal ids
                o = self.context[value]

                #if key == 'source':
                    #yield o, readable.sourceOf, self.s
                #elif key == 'target':
                    #yield o, readable.targetOf, self.s

                yield self.s, readable[key], o

    def triples_objects_multi(self):
        for key in self.objects_multi:
            if key in self.blob:
                values = self.blob[key]
                assert not isinstance(values, str), f'{values} in {key}'
                for value in values:
                    if key == 'external':
                        try:
                            o = OntId(value).URIRef
                            yield o, readable.annotates, self.s
                        except OntId.UnknownPrefixError as e:
                            log.exception(e)
                            continue
                    elif key == 'inheritedExternal':
                        try:
                            o = OntId(value).URIRef
                        except OntId.UnknownPrefixError as e:
                            log.exception(e)
                            continue
                    else:
                        value = value.replace(' ', '-')  # FIXME require no spaces in internal ids
                        o = self.context[value]

                    yield self.s, readable[key], o

    def triples_objects_ordered(self):
        for key, predicate in self.objects_ordered_succession.items():
            if key in self.blob:
                values = self.blob[key]
                if values:
                    assert not isinstance(values, str), f'{values} in {key}'
                    objects = [OntId(self.context[v.replace(' ', '-')]).URIRef for v in values]
                    for s, o in zip(objects[:-1],objects[1:]):
                        yield s, predicate, o

        for key in self.objects_ordered:
            if key in self.blob:
                values = self.blob[key]
                if values:
                    assert not isinstance(values, str), f'{values} in {key}'
                    objects = [OntId(self.context[v.replace(' ', '-')]).URIRef for v in values]
                    yield from cmb.olist(*objects)(self.s, ordered[key])  # NOTE scigraph does not translate rdf lists

    def triples_external(self):
        if 'externals' in self.blob:
            for external in self.blob['external']:
                yield self.s, rdf.type, OntId(external).URIRef


class Node(BaseElement):
    key = 'nodes'
    annotations = 'skipLabel', 'color', 'generated'
    objects = 'cloneOf', 'hostedBy', 'internalIn',
    objects_multi = 'sourceOf', 'targetOf', 'clones', 'external', 'rootOf', 'leafOf',

    def triples(self):
        yield from super().triples()


Graph.Node = Node
class Lyph(BaseElement):
    key = 'lyphs'
    generics = 'topology',
    annotations = 'width', 'height', 'layerWidth', 'internalLyphColumns', 'isTemplate', 'generated'
    objects = 'layerIn', 'conveys', 'border', 'cloneOf', 'supertype', 'internalIn'
    objects_multi = ('inCoalescences', 'subtypes', 'layers', 'clones', 'external', 'inheritedExternal'
                     'internalNodes', 'bundles', 'bundlesTrees', 'bundlesChains', 'subtypes')

    def triples(self):
        yield from super().triples()


Graph.Lyph = Lyph
class Link(BaseElement):
    key = 'links'
    annotations = 'generated',
    generics = 'conveyingType',
    objects = 'source', 'target', 'conveyingLyph', 'fasciculatesIn'
    objects_multi = 'conveyingMaterials', 'hostedNodes', 'external', 'inheritedExternal',  'next', 'prev'


Graph.Link = Link
class Coalescence(BaseElement):
    key = 'coalescences'
    generics = 'topology',
    annotations = 'generated',
    objects = 'generatedFrom',
    objects_multi = 'lyphs', 'external', 'inheritedExternal'


Graph.Coalescence = Coalescence
class Border(BaseElement):
    # FIXME class is Link ?
    key = 'borders'
    objects = 'host',
    objects_multi = 'borders', 'external', 'inheritedExternal'


Graph.Border = Border
class Tree(BaseElement):
    key = 'trees'
    objects = 'root', 'lyphTemplate', 'group'
    objects_multi = 'housingLyphs', 'external', 'levels', 'inheritedExternal'


Graph.Tree = Tree
class Chain(BaseElement):
    key = 'chains'
    #internal_references = 'housingLayers',   # FIXME TODO
    objects = 'root', 'leaf', 'lyphTemplate', 'group', 'housingChain'
    objects_multi = 'housingLyphs', 'external', 'levels', 'lyphs', 'inheritedExternal'
    #objects_ordered_succession = {'lyphs': readable.nextLyph,
                                  #'levels': readable['next']}


Graph.Chain = Chain
class Group(BaseElement):
    key = 'groups'
    objects = 'generatedFrom',
    objects_multi = 'nodes', 'links', 'Lyphs', 'coalescences', 'groups'
    elements = Node, Link, Lyph, Coalescence  # Group  # ah class scope

    def triples(self):
        yield from super().triples()
        for element_class in self.elements:
            if element_class.key in self.blob:
                for value_local_id in self.blob[element_class.key]:
                    c = element_class({'id':value_local_id}, context=self.context)
                    yield self.s, readable.hasElement, c.s
            else:
                log.warning(f'{element_class.key} not in {self.blob}')


Group.elements += (Group,)
Graph.Group = Group
class Material(BaseElement):
    key = 'materials'
    objects_multi = 'materials', 'inMaterials', 'external', 'inheritedExternal'


class fake:
    """ filler """
    label = 'sigh'


Graph.Material = Material
class External(BaseElement):
    externals = 'id',
    annotations = 'generated', 'uri', 'type',  # FIXME should be classes
    objects_multi = 'annotates',

    @classmethod
    def fromRdf(cls, uri, graph, context=None):
        oid = OntId(uri)
        id = oid.curie
        blob = {'id': id}
        for p, o in graph[uri]:
            if p == rdf.type:
                key = 'class'
                value = 'External'
            else:
                if p == rdfs.label:
                    key = 'name'
                else:
                    _, key = p.rsplit('/', 1)

                if isinstance(o, rdflib.Literal):
                    value = o.toPython()
                elif isinstance(o, rdflib.URIRef):
                    oid = OntId(o)
                    if oid.prefix == 'local':
                        value = oid.suffix
                    elif oid.prefix == 'apinatomy':  # FIXME hrm?
                        value = oid.suffix
                    else:
                        value = oid.curie  # FIXME external is tricky
                        log.warning(f'{oid!r}')

            if key in cls.objects_multi:
                if key in blob:
                    blob[key].append(value)
                else:
                    blob[key] = [value]

            else:
                blob[key] = value
            
        return cls(blob, context)

    @mimicArgs(BaseElement.__init__)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # PLEASE DO NOT PUT PMIDs as external ids!!!
        # FIXME idlib PMID(thing) urg the regex state machine is so simple ;_;
        if self.id.startswith('PMID:'):
            log.warning('PMIDs should never be External IDs!')
            self._c_term = fake
            self.s = OntId(self.id).URIRef
            return

        self.s = OntId(self.id).URIRef

    @property
    def _term(self):
        if not hasattr(self, '_c_term'):
            self._c_term = OntTerm(self.id)

        return self._c_term

    def triples_resource(self):
        yield self.s, rdf.type, owl.Class
        l = self._term.label if self._term.label is not None else self._term.curie
        yield self.s, rdfs.label, rdflib.Literal(l)
        # TODO triples simple?


Graph.External = External
class Channel(BaseElement):
    pass


Graph.Channel = Channel


if __name__ == '__main__':
    # FIXME this eternal annoyance of dealing with the network at top level
    # of course if you want to generate some code based on something from the
    # network then of course you have to do this at some point otherwise
    # you have to figure out how to defer loading until the last possible moment
    apinscm = sc.ApiNATOMYSchema()
    hrm = make_classes(apinscm.schema)



import json
import itertools
from collections import defaultdict
from urllib.parse import urlparse, parse_qs
import rdflib
import requests
import augpathlib as aug
from pyontutils.config import auth as pauth
from pyontutils.core import OntId
from pyontutils.utils import byCol
from sparcur import normalization as nml
from sparcur.core import log, logd, JEncode
from sparcur.paths import Path
#from sparcur.utils import cache
from sparcur.config import config, auth
from bs4 import BeautifulSoup

# ontology files
class OntologyData:

    def _mis_graph(self):
        """ for now easier to just get a fresh one, they are small """
        glb = pauth.get_path('git-local-base')
        olr = Path(glb / 'duplicates' / 'sparc-NIF-Ontology')
        graph = (rdflib.ConjunctiveGraph()
            .parse((olr / 'ttl/sparc-methods.ttl').as_posix(), format='turtle')
            #.parse((olr / 'ttl/methods-core.ttl').as_posix(), format='turtle')
            #.parse((olr / 'ttl/methods-helper.ttl').as_posix(), format='turtle')
            #.parse((olr / 'ttl/methods.ttl').as_posix(), format='turtle')
        )
        return graph

    def run_reasoner(self):
        graph = self._mis_graph()
        expanded_graph = self._mis_graph()
        [(graph.add(t), expanded_graph.add(t))
         for t in self.triples()]
        closure = rdfc.OWLRL_Semantics
        rdfc.DeductiveClosure(closure).expand(expanded_graph)
        with open(auth.get_path('cache-path') / 'reasoned-curation-export.ttl', 'wb') as f:
            f.write(expanded_graph.serialize(format='nifttl'))


# other

class OrganData:
    """ retrieve SPARC investigator data """

    url = ('https://commonfund.nih.gov/sites/default/'
           'files/sparc_nervous_system_graphic/main.html')

    def organ(self, award_number):
        if award_number in self.manual and award_number not in self.sourced:
            log.warning(f'used manual organ mapping for {award_number}')
        try:
            return self.award_to_organ[award_number]
        except KeyError as e:
            logd.error(f'bad award_number {award_number}')

    __call__ = organ

    organ_lookup = {'bladder': OntId('FMA:15900'),
                    'brain': OntId('UBERON:0000955'),
                    #'computer': OntId(''),
                    'heart': OntId('FMA:7088'),
                    'kidneys': OntId('FMA:7203'),
                    'largeintestine': OntId('FMA:7201'),
                    'liver': OntId('FMA:7197'),
                    'lung': OntId('FMA:7195'),
                    'malerepro': OntId('UBERON:0000079'),
                    #'othertargets': OntId(''),
                    'pancreas': OntId('FMA:7198'),
                    'smallintestine': OntId('FMA:7200'),
                    'spleen': OntId('FMA:7196'),
                    'stomach': OntId('FMA:7148'),
                    'vagus nerve': OntId('FMA:5731'),
                    #'uterus': OntId('')
                    '': None,
    }

    cache = auth.get_path('cache-path') / 'sparc-award-by-organ.json'
    old_cache = auth.get_path('cache-path') / 'award-mappings-old-to-new.json'

    def __init__(self, path=config.organ_html_path, organs_sheet=None):  # FIXME bad passing in organs
        self.path = path
        if not self.cache.exists():
            self.overview()
            with open(self.cache, 'wt') as f:
                json.dump(self.normalized, f)

            with open(self.old_cache, 'wt') as f:
                json.dump(self.former_to_current, f)
        else:
            with open(self.cache, 'rt') as f:
                self.normalized = json.load(f)

            with open(self.old_cache, 'rt') as f:
                self.former_to_current = json.load(f)

        if organs_sheet is not None:
            self._org = organs_sheet
            bc = self._org.byCol
            self.manual  = {award if award else (award_manual if award_manual else None):[OntId(t) for t in organ_term.split(' ') if t]
                            for award, award_manual, organ_term
                            in zip(bc.award, bc.award_manual, bc.organ_term)
                            if organ_term}
        else:
            self.manual = {}

        self.sourced = {v:k for k, vs in self.normalized.items() for v in vs}
        self.award_to_organ = {**self.sourced, **self.manual}  # manual override

    def overview(self):
        if self.path.exists():
            with open(self.path, 'rb') as f:
                soup = BeautifulSoup(f.read(), 'lxml')
        else:
            resp = requests.get(self.url)
            soup = BeautifulSoup(resp.content, 'lxml')

        self.raw = {}
        self.former_to_current = {}
        for bsoup in soup.find_all('div', {'id':lambda v: v and v.endswith('-bubble')}):
            organ, *_rest = bsoup['id'].split('-')
            logd.debug(_rest)
            award_list = self.raw[organ] = []
            for asoup in bsoup.find_all('a'):
                href = asoup['href']
                log.debug(href)
                parts = urlparse(href)
                query = parse_qs(parts.query)
                if 'projectnumber' in query:
                    award_list.extend(query['projectnumber'])
                elif 'aid' in query:
                    #aid = [int(a) for a in query['aid']]
                    #json = self.reporter(aid)
                    award, former = self.reporter(href)
                    award_list.append(award)
                    if former is not None:
                        award_list.append(former)  # for this usecase this is ok
                        self.former_to_current[former] = award
                elif query:
                    log.debug(lj(query))
            
        self.former_to_current = {nml.NormAward(nml.NormAward(k)):nml.NormAward(nml.NormAward(v))
                                  for k, v in self.former_to_current.items()}
        self._normalized = {}
        self.normalized = {}
        for frm, to in ((self.raw, self._normalized), (self._normalized, self.normalized)):
            for organ, awards in frm.items():
                if organ in self.organ_lookup:
                    organ = self.organ_lookup[organ].iri

                to[organ] = [nml.NormAward(a) for a in awards]

    def _reporter(self, aids):
        # can't seem to get this to cooperate
        base = ('https://api.federalreporter.nih.gov'
                '/v1/projects/FetchBySmApplIds')
        resp = requests.post(base, json=aids, headers={'Accept': 'application/json',
                                                       'Content-Type': 'application/json'})
        breakpoint()
        return resp.json()

    def reporter(self, href):
        resp = requests.get(href)
        soup = BeautifulSoup(resp.content, 'lxml')
        #id = soup.find_all('span', {'id': 'spnPNUMB'})
        table = soup.find_all('table', {'summary': 'Details'})
        text = table[0].find_all('td')[1].text.strip()
        if 'Former' in text:
            award, rest = text.split(' ', 1)
            rest, former = text.rsplit(' ', 1)
            return [award, former]
        else:
            return [text, None]


class MembersData:
    """ blackfynn members data """
    def __init__(self, blackfynn_local_instance):
        self.__class__._bfli = blackfynn_local_instance

    def __call__(self, first_name, last_name):
        return self.get_member_by_name(first_name, last_name)

    @property
    def members(self):
        if not hasattr(self.__class__, '_members'):
            self.__class__._members = self._bfli.organization.members

        return self._members

    @property
    def __members(self):
        if not hasattr(self.__class__, '_members'):
            log.debug('going to network for members')
            # there are other ways to get here but this one caches
            # e.g. self.organization.path.remote.bfobject
            # self.path.remote.oranization.bfobject
            # self.path.remote.bfl.organization.members
            self.__class__._members = self.path.remote.bfl.organization.members

        return self._members

    @property
    def member_index_f(self):
        if not hasattr(self.__class__, '_member_index_f'):
            mems = defaultdict(lambda:defaultdict(list))
            for member in self.members:
                fn = member.first_name.lower()
                ln = member.last_name.lower()
                current = mems[fn][ln].append(member)

            self.__class__._member_index_f = {fn:dict(lnd) for fn, lnd in mems.items()}

        return self._member_index_f

    @property
    def member_index_l(self):
        if not hasattr(self.__class__, '_member_index_l'):
            mems = defaultdict(dict)
            for fn, lnd in self.member_index_f.items():
                for ln, member_list in lnd.items():
                    mems[ln][fn] = member_list

            self.__class__._member_index_l = dict(mems)

        return self._member_index_l

    def get_member_by_name(self, first, last):
        def lookup(d, one, two):
            if one in d:
                ind = d[one]
                if two in ind:
                    member_list = ind[two]
                    if member_list:
                        member = member_list[0]
                        if len(member_list) > 1:
                            log.critical(f'WE NEED ORCIDS! {one} {two} -> {member_list}')
                            # organization maybe?
                            # or better, check by dataset?

                        return member

        fnd = self.member_index_f
        lnd = self.member_index_l
        fn = first.lower()
        ln = last.lower()
        m = lookup(fnd, fn, ln)
        if not m:
            m = lookup(lnd, ln, fn)

        return m


class BlackfynnDatasetData:

    cache_path = auth.get_path('cache-path')
    cache_base = cache_path / 'blackfynn-meta'

    def __init__(self, remote_cache_or_id):
        if isinstance(remote_cache_or_id, aug.RemotePath):
            self._remote = remote_cache_or_id
            self._bfobject = self._remote.bfobject
            self.id = self._remote.id
        elif isinstance(remote_cache_or_id, aug.CachePath):
            self._c_cache_path = remote_cache_or_id
            self.id = self._c_cache_path.id
        else:
            self.id = remote_cache_or_id

        self.ntfs_safe_id = self.id.split(':')[-1]  # sigh
        self.cache = self.cache_base / self.ntfs_safe_id
        if not self.cache_base.exists():
            self.cache_base.mkdir()

    @property
    def _cache_path(self):
        if not hasattr(self, '_c_cache_path'):
            raise NotImplementedError('if you need access to the cache '
                                      'pass one in at construction time')

        return self._c_cache_path

    @property
    def remote(self):
        if not hasattr(self, '_remote'):
            self._remote = self._cache_path.remote

        return self._remote

    @property
    def bfobject(self):
        if not hasattr(self, '_bfobject'):
            self._bfobject = self.remote.bfobject

        return self._bfobject

    def fromCache(self):
        """ retrieve cached results without hitting the network """
        # TODO FIXME error on no cache?
        if not self.cache.exists():
            msg = f'No cached metadata for {self.id}. Run `spc rmeta` to populate.'
            raise FileNotFoundError(msg)

        with open(self.cache, 'rt') as f:
            return json.load(f)

    def __call__(self):
        # FIXME TODO switch to use dict transformers
        # self.bfobject.relationships()
        meta = self.bfobject.meta  # why this is not default in the python api the world may never know
        package_counts = self.bfobject.packageTypeCounts
        cont = meta['content']
        blob = {'id': self.remote.id,
                'name': cont['name'],  # title
                'description': cont['description'],  # subtitle
                'readme': self.bfobject.readme,
                # 'banner': self.bfobject.banner,  # FIXME not persistent ...
                'status-log': self.bfobject.status_log,  # FIXME most recent only?
                'tags': cont['tags'],

                # FIXME strip emails before export
                # leave teams and users out of this for now, too much noise
                #'teams': self.bfobject.teams,
                #'users': self.bfobject.users,
                'contributors': self.bfobject.contributors,
                'package_counts': package_counts,
        }

        if 'license' in cont:
            blob['license'] = cont['license']

        doi = self.remote.doi
        if doi is not None:
            blob['doi'] = self.remote.doi

        with open(self.cache, 'wt') as f:
            json.dump(blob, f, indent=2, sort_keys=True, cls=JEncode)

        return blob



import json
import unittest
from pathlib import Path
from sparcur import apinat
import pytest

export = False
debug = False


class TestApiNATToRDF(unittest.TestCase):
    def test_1(self):
        with open((Path(__file__).parent / 'apinatomy/test_1_map.json'), 'rt') as f:
            m = json.load(f)

        apin = apinat.Graph(m)
        rdfg = apin.graph()
        rtmap = apinat.Graph.fromRdf(rdfg).map
        assert list(rdfg)
        assert rtmap['resources']
        if export:
            rdfg.write(Path('test-1.ttl'))
        if debug:
            print(rdfg.ttl)
            print(rtmap)
            assert False

    def test_2(self):
        with open((Path(__file__).parent / 'apinatomy/test_2_map.json'), 'rt') as f:
            m = json.load(f)

        apin = apinat.Graph(m)
        rdfg = apin.graph()
        rtmap = apinat.Graph.fromRdf(rdfg).map
        assert list(rdfg)
        assert rtmap['resources']
        if export:
            rdfg.write(Path('test-2.ttl'))
        if debug:
            print(rdfg.ttl)
            print(rtmap)
            assert False

    def test_3(self):
        with open((Path(__file__).parent / 'apinatomy/test_3_map.json'), 'rt') as f:
            m = json.load(f)

        apin = apinat.Graph(m)
        rdfg = apin.graph()
        rtmap = apinat.Graph.fromRdf(rdfg).map
        assert list(rdfg)
        assert rtmap['resources']
        if export:
            rdfg.write(Path('test-3.ttl'))
        if debug:
            print(rdfg.ttl)
            print(rtmap)
            assert False

    def test_4(self):
        with open((Path(__file__).parent / 'apinatomy/test_4_map.json'), 'rt') as f:
            m = json.load(f)

        apin = apinat.Graph(m)
        rdfg = apin.graph()
        rtmap = apinat.Graph.fromRdf(rdfg).map
        assert list(rdfg)
        assert rtmap['resources']
        if export:
            rdfg.write(Path('test-4.ttl'))
        if debug:
            print(rdfg.ttl)
            print(rtmap)
            assert False

    def test_5(self):
        with open((Path(__file__).parent / 'apinatomy/test_5_map.json'), 'rt') as f:
            m = json.load(f)

        apin = apinat.Graph(m)
        rdfg = apin.graph()
        rtmap = apinat.Graph.fromRdf(rdfg).map
        assert list(rdfg)
        assert rtmap['resources']
        if export:
            rdfg.write(Path('test-5.ttl'))
        if debug:
            print(rdfg.ttl)
            print(rtmap)
            assert False

    def test_bolew(self):
        with open((Path(__file__).parent / 'apinatomy/bolser-lewis-map.json'), 'rt') as f:
            m = json.load(f)

        apin = apinat.Graph(m)
        rdfg = apin.graph()
        if export:
            rdfg.write(Path('test-bolew.ttl'))
        rtmap = apinat.Graph.fromRdf(rdfg).map
        if debug:
            print(rdfg.ttl)
            print(rtmap)
            assert False

    def test_keast(self):
        with open((Path(__file__).parent / 'apinatomy/keast-spinal-map.json'), 'rt') as f:
            m = json.load(f)

        apin = apinat.Graph(m)
        rdfg = apin.graph()
        if export:
            rdfg.write(Path('keast-spinal.ttl'))
        rtmap = apinat.Graph.fromRdf(rdfg).map
        if debug:
            print(rdfg.ttl)
            print(rtmap)
            assert False


class TestRDFToOWL2(unittest.TestCase):
    pass
